__description__ = \
"""
An Enigma simulator and analyser

Versions:
 1.0.0      first version
 1.1.0      documentation established
 1.1.2      setup corrected
"""
__author__ = "Reinhard Maerz"
__date__ = "2022-03-26"
__version__ = "1.1.2"

import os
import sys

sys.path.append(os.path.dirname(os.path.abspath(__file__)))

__all__ = [
 'Umkehrwalze', 'Steckerbrett', 'Zusatzwalze', 'Walze', 
 'stdAlphabet', 'sgsAlphabet', 
 'SpruchScoring', 'Enigma', 'Tagesschluessel', 'TagesschluesselRange'
]

from .component import Umkehrwalze, Steckerbrett, Zusatzwalze, Walze, Component
from .component import stdAlphabet, sgsAlphabet
from .enigma import SpruchScoring, Enigma, Tagesschluessel
from .analyzeEnigma import TagesschluesselRange

############# Components ############# 

# Steckerbrett
'''
for preconfigured Steckerbretts, see `MARK`_1, `MARK`_2, `MARK`_3
.. _MARK:
'''
MARK_1 = Steckerbrett.Mark_1() 
MARK_2 = Steckerbrett.Mark_2() 
MARK_3 = Steckerbrett.Mark_3(nConnections = 10)  
UnconnectedSteckerbrett = Steckerbrett('Unconnected', stdAlphabet) 

# 1924 Commercial Enigma A, B Mark 1
I_A = Walze(name = 'I_A', wiring = 'DMTWSILRUYQNKFEJCAZBPGXOHV', notches = 'Y') 
II_A = Walze(name = 'II_A', wiring = 'HQZGPJTMOBLNCIFDYAWVEUSRKX', notches = 'Y') 
III_A = Walze(name = 'III_A', wiring = 'UQNTLSZFMREHDPXKIBVYGJCWOA', notches = 'Y') 
UKW_D = Umkehrwalze(name = 'UKW_D', wiring = 'IMETCGFRAYSQBZXWLHKDVUPOJN') 

Enigma_A = Enigma(model = 'Enigma A', walzen = [I_A, II_A, III_A], umkehrwalzen = [UKW_D], numberOfWalzen = 2)

# 1925 Commercial Enigma A-133
I_SGS = Walze(name = 'I_SGS', wiring = 'PSBGÖXQJDHOÄUCFRTEZVÅINLYMKA', notches = 'Ä', alphabet = sgsAlphabet)
II_SGS = Walze(name = 'II_SGS', wiring = 'CHNSYÖADMOTRZXBÄIGÅEKQUPFLVJ', notches = 'Ä', alphabet = sgsAlphabet)
III_SGS = Walze(name = 'III_SGS', wiring = 'ÅVQIAÄXRJBÖZSPCFYUNTHDOMEKGL', notches = 'Ä', alphabet = sgsAlphabet)
UKW_SGS = Umkehrwalze(name = 'UKW_SGS', wiring = 'LDGBÄNCPSKJAVFZHXUIÅRMQÖOTEY', alphabet = sgsAlphabet)

Enigma_SGS = Enigma(model = 'Enigma SGS', walzen = [I_SGS, II_SGS, III_SGS], umkehrwalzen = [UKW_SGS], numberOfWalzen = 2)

# 1927 Reichswehr Enigma D, K, Commercial Enigma A26 and A28
I_D = Walze(name = 'I_D', wiring = 'LPGSZMHAEOQKVXRFYBUTNICJDW', notches = 'Y')
II_D = Walze(name = 'II_D', wiring = 'SLVGBTFXJQOHEWIRZYAMKPCNDU', notches = 'E')
III_D = Walze(name = 'III_D', wiring = 'CJGDPSHKTURAWZXFMYNQOBVLIE', notches = 'N')

Enigma_D = Enigma(model = 'Enigma D', walzen = [I_D, II_D, III_D], umkehrwalzen = [UKW_D], numberOfWalzen = 3)

# 1929, Reichswehr Enigma MI
I = Walze(name = 'I', wiring = 'EKMFLGDQVZNTOWYHXUSPAIBRCJ', notches = 'Q')
II = Walze(name = 'II', wiring = 'AJDKSIRUXBLHWTMCQGZNPYFVOE', notches = 'E')
III = Walze(name = 'III', wiring = 'BDFHJLCPRTXVZNYEIWGAKMUSQO', notches = 'V')
IV = Walze(name = 'IV', wiring = 'ESOVPZJAYQUIRHXLNFTGKDCMWB', notches = 'J')
V = Walze(name = 'V', wiring = 'VZBRGITYUPSDNHLXAWMJQOFECK', notches = 'Z')
UKW_A = Umkehrwalze(name = 'UKW-A', wiring = 'EJMZALYXVBWFCRQUONTSPIKHGD')	 	 	 
UKW_B = Umkehrwalze(name = 'UKW-B', wiring = 'YRUHQSLDPXNGOKMIEBFZCWVJAT')	 	 	 
UKW_C = Umkehrwalze(name = 'UKW-C', wiring = 'FVPJIAOYEDRZXWGCTKUQSBNMHL')

Enigma_I = Enigma(model = 'Enigma I', walzen = [I, II, III, IV, V], umkehrwalzen = [UKW_A, UKW_B, UKW_C], steckerbrett = MARK_3, numberOfWalzen = 3)

# 1940, Kriegsmarine Enigma M3
VI = Walze(name = 'VI', wiring = 'JPGVOUMFYQBENHZRDKASXLICTW', notches = 'ZM')
VII = Walze(name = 'VII', wiring = 'NZJHGRCXMYSWBOUFAIVLPEKQDT', notches = 'ZM')
VIII = Walze(name = 'VIII', wiring = 'FKQHTLXOCBJSPDZRAMEWNIUYGV', notches = 'ZM')

Enigma_M3 = Enigma(model = 'Enigma M3', walzen = [I, II, III, IV, V, VI, VII, VIII], 
                             umkehrwalzen = [UKW_A, UKW_B, UKW_C], steckerbrett = MARK_3, numberOfWalzen = 3)

# 1941, U-Boot Enigma M4
Beta = Zusatzwalze(name = 'Beta', wiring = 'LEYJVCNIXWPBQMDRTAKZGFUHOS') 	 	 
Gamma = Zusatzwalze(name = 'Gamma', wiring = 'FSOKANUERHMBTIYCWLQPZXVGJD') 	 
UKW_ThinB = Umkehrwalze(name = 'UKW-ThinB',	wiring = 'ENKQAUYWJICOPBLMDXZVFTHRGS') 
UKW_ThinC = Umkehrwalze(name = 'UKW-ThinC',	wiring = 'RDOBJNTKVEHMLFCWZAXGYIPSUQ')

Enigma_M4 = Enigma(model = 'Enigma M4', walzen = [I, II, III, IV, V, VI, VII, VIII], umkehrwalzen = [UKW_ThinB, UKW_ThinC], 
                             steckerbrett = MARK_3, zusatzwalzen = [Beta, Gamma],  numberOfWalzen = 4)

# 1942 Enigma T (Tirpitz)
I_T 	= Walze(name = 'I_T', wiring = 'KPTYUELOCVGRFQDANJMBSWHZXI', notches = 'WZEKQ')
II_T 	= Walze(name = 'II_T', wiring = 'UPHZLWEQMTDJXCAKSOIGVBYFNR', notches = 'WZFLR')
III_T 	= Walze(name = 'III_T', wiring = 'QUDLYRFEKONVZAXWHMGPJBSICT', notches = 'WZEKQ')
IV_T 	= Walze(name = 'IV_T', wiring = 'CIWTBKXNRESPFLYDAGVHQUOJZM', notches = 'WZFLR')
V_T 	= Walze(name = 'V_T', wiring = 'UAXGISNJBVERDYLFZWTPCKOHMQ', notches = 'YCFKR')
VI_T 	= Walze(name = 'VI_T', wiring = 'XFUZGALVHCNYSEWQTDMRBKPIOJ', notches = 'XEIMQ')
VII_T 	= Walze(name = 'VII_T', wiring = 'BJVFTXPLNAYOZIKWGDQERUCHSM', notches = 'YCFKR')
VIII_T 	= Walze(name = 'VIII_T', wiring = 'YMTPNZHWKODAJXELUQVGCBISFR', notches = 'XEIMQ')
UKW_T = Umkehrwalze(name = 'UKW_T',	wiring = 'GEKPBTAUMOCNILJDXZYFHWVQSR') 

Enigma_T = Enigma(model = 'Enigma Tirpitz', walzen = [I_T, II_T, III_T, IV_T, V_T, VI_T, VII_T, VIII_T], 
                           umkehrwalzen = [UKW_T], numberOfWalzen = 3)

# 1931 G31 Abwehr Enigma
I_G312 = Walze(name = 'I_G312', wiring = 'DMTWSILRUYQNKFEJCAZBPGXOHV', notches = 'SUVWZABCEFGIKLOPQ')
II_G312 = Walze(name = 'II_G312', wiring = 'HQZGPJTMOBLNCIFDYAWVEUSRKX', notches = 'STVYZACDFGHKMNQ')
III_G312 = Walze(name = 'III_G312', wiring = 'UQNTLSZFMREHDPXKIBVYGJCWOA', notches = 'UWXAEFHKMNR')
UKW_G312 = Umkehrwalze(name = 'UKW_G312', wiring = 'RULQMZJSYGOCETKWDAHNBXPVIF')

Enigma_G312 = Enigma(model = 'Enigma G-312', walzen = [I_G312, II_G312, III_G312], 
                           umkehrwalzen = [UKW_G312], numberOfWalzen = 3)

# 1945 Norway Enigma (Norenigma)
I_N = Walze(name = 'I_N', wiring = 'WTOKASUYVRBXJHQCPZEFMDINLG', notches = 'Q')
II_N = Walze(name = 'II_N', wiring = 'GJLPUBSWEMCTQVHXAOFZDRKYNI', notches = 'E')
III_N = Walze(name = 'III_N', wiring = 'JWFMHNBPUSDYTIXVZGRQLAOEKC', notches = 'V')
IV_N = Walze(name = 'IV_N', wiring = 'FGZJMVXEPBWSHQTLIUDYKCNRAO', notches = 'J')
V_N = Walze(name = 'V_N', wiring = 'HEJXQOTZBVFDASCILWPGYNMURK', notches = 'Z')
UKW_N = Umkehrwalze(name = 'UKW_N', wiring = 'MOWJYPUXNDSRAIBFVLKZGQCHET')

Enigma_N = Enigma(model = 'Norenigma', walzen = [I_N, II_N, III_N, IV_N, V_N], 
                           umkehrwalzen = [UKW_N], numberOfWalzen = 3)




# from typing import Dict, Tuple, Any
import pytest

import copy
import random
import re
import MzEnigma

notify = [None, print]

def setWorstTagesWalzenStellungen(tagesschluessel : MzEnigma.Tagesschluessel) -> None:
 tagesWalzenStellungen = str()
 for walze in tagesschluessel.walzen:
  tagesWalzenStellungen += walze.notches[0]
 tagesschluessel.tagesWalzenStellungen = tagesWalzenStellungen

def test_print(pytestconfig):
 pattern = pytestconfig.getoption('component')
 if pattern:
  print('\n--- test_print: pattern = {} ---'.format(pattern))
  for name, cls in vars(MzEnigma).items():
   if (isinstance(cls, MzEnigma.Umkehrwalze) \
    or isinstance(cls, MzEnigma.Enigma) \
    or isinstance(cls, MzEnigma.Tagesschluessel)) \
    and re.fullmatch(pattern, name):
    print('\n -{}'.format(cls))

def test_basicencode(pytestconfig):
 print('\n--- test_basicencode ---')
 enigma = MzEnigma.Enigma(model = 'Enigma Test', walzen = [MzEnigma.I, MzEnigma.II], umkehrwalzen = [MzEnigma.UKW_B], numberOfWalzen = 2)
 msg = enigma.alphabet[:2]
 print(enigma)
 enigmaSetting = MzEnigma.Tagesschluessel(enigma, walzen = [MzEnigma.I, MzEnigma.II], tagesWalzenStellungen = 'AA', notify = pytest.helpers.notify(pytestconfig))
 # chainList = enigmaSetting.chain()
 eMsg = enigmaSetting.encode(msg)
 newMsg = enigmaSetting.decode(eMsg)
 assert msg == newMsg, 'Encoding of msg {} failed on enigma {}'.format(msg, enigma.model)

def test_chain(pytestconfig):
 msg = pytest.helpers.message(pytestconfig)
 pattern = pytestconfig.getoption('component')
 if pattern:
  print('\n--- test_chain ---\nmessage: {}'.format(msg))
  for name, enigma in vars(MzEnigma).items():
   if isinstance(enigma, MzEnigma.Enigma) and re.fullmatch(pattern, name):
    enigmaSetting = MzEnigma.Tagesschluessel(enigma, notify = pytest.helpers.notify(pytestconfig))
    enigmaSetting.chain()
    print('- {} completed'.format(name))

def test_encode(pytestconfig):
 msg = pytest.helpers.message(pytestconfig)
 pattern = pytestconfig.getoption('component')
 if pattern:
  print('\n--- test_encode ---\n')
  for name, enigma in vars(MzEnigma).items():
   if isinstance(enigma, MzEnigma.Enigma) and re.fullmatch(pattern, name):
    enigmaSetting = MzEnigma.Tagesschluessel(enigma, notify = pytest.helpers.notify(pytestconfig))
    setWorstTagesWalzenStellungen(enigmaSetting)
    print(enigmaSetting)
    if msg:
     actMsg = msg
    else:
     actMsg = enigma.alphabet
    eMsg = enigmaSetting.encode(actMsg)
    newMsg = enigmaSetting.decode(eMsg)
    assert actMsg == newMsg, 'Encoding of msg {} failed on enigma {}'.format(actMsg, enigma.model)
    print('- {} completed'.format(name))

def test_encode2(pytestconfig):
 msg = pytest.helpers.message(pytestconfig)
 pattern = pytestconfig.getoption('component')
 if pattern:
  print('\n--- test_encode2 (using Spruchschluessel) ---\n')
  for name, enigma in vars(MzEnigma).items():
   if isinstance(enigma, MzEnigma.Enigma) and re.fullmatch(pattern, name):
    enigmaSetting = MzEnigma.Tagesschluessel(enigma, notify = pytest.helpers.notify(pytestconfig))
    setWorstTagesWalzenStellungen(enigmaSetting)
    print(enigmaSetting)
    if msg:
     actMsg = msg
    else:
     actMsg = enigma.alphabet    
    eMsg = enigmaSetting.encode(actMsg, spruchWalzenStellungen = ''.join(random.sample(actMsg, enigma.numberOfWalzen)))
    newMsg = enigmaSetting.decode(eMsg, useSpruchWalzenStellungen = True)
    print('{} -> {} -> {}'.format(msg, eMsg,  newMsg))
    assert actMsg == newMsg, 'Encoding of msg {} failed on enigma {}'.format(actMsg, enigma.model)
    print('- {} completed'.format(name))

if __name__ == "__main__":
 pattern = '.+'
 for name, enigma in vars(MzEnigma).items():
  if isinstance(enigma, MzEnigma.Enigma) and re.fullmatch(pattern, name):
   enigmaSetting = MzEnigma.Tagesschluessel(enigma,  notify = None)
   spruchWalzenStellungen = ''.join(random.sample(enigma.alphabet, enigma.numberOfWalzen))
   twiceEncodedSpruchWalzenStellungen = enigmaSetting.encode(spruchWalzenStellungen + spruchWalzenStellungen)
   print('{}/{} -> {}: decoded = {}'.format(enigmaSetting.tagesWalzenStellungen, spruchWalzenStellungen + spruchWalzenStellungen, twiceEncodedSpruchWalzenStellungen, enigmaSetting.decode(twiceEncodedSpruchWalzenStellungen)))
   break




from typing import TypeVar,  Union, Optional, Callable, Dict,  List,  Tuple
import random
import copy

stdAlphabet : str = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
sgsAlphabet : str = 'ABCDEFGHIJKLMNOPQRSTUVXYZÅÄÖ'

class Umkehrwalze: pass
class Steckerbrett: pass
class Zusatzwalze: pass
class Walze: pass
Component = TypeVar('Component', Umkehrwalze, Steckerbrett, Zusatzwalze, Walze)

class Umkehrwalze(object):
 """Represents an Umkehrwalze (UKW == reflector)

.. csv-table:: Wheel Controls
   :header: "Item", "Value"
   :widths: 30, 30
   
   *encryption*, forward only 
   *ring*, none
   *notches*, none 

:param name: name of the Umkehrwalze
:param wiring: wiring of the Umkehrwalze, positions related to the *alphabet*
:param alphabet: unencoded alphabet to be used
:param notify: notification function (e.g. print)
 """
 def __init__(self, 
  name  : str = '', 
  wiring : Union[str, Dict[str, str]] = '', 
  alphabet  : str = stdAlphabet, 
  notify : Optional[Callable[[str], None]] = None) -> None:
  self.notify = notify
  self._nextComponent : Optional[Component]  = None
  self._prevComponent : Optional[Component] = None
  assert name, '{}: At least 1 character in name required'.format(self.__class__.__name__)
  self._name = name
  assert alphabet, '{}/{}: At least 1 character in alphabet required'.format(self.__class__.__name__, self._name)
  self._alphabet = alphabet
  self._numberOfPositions = len(self.alphabet)
 
  if isinstance(wiring, dict):
   wiringList = self.alphabet.split('')
   for src, tgt in wiring.items():
    assert src in self.alphabet, '{}: Source character {} not in {}'.format(self.__class__.__name__, src, self.alphabet)
    srcID = self.alphabet.index(src)
    wiringList[srcID] = tgt
   wiring = ''.join(wiringList)
  self.wiring = wiring
  if self.__class__.__name__ == 'Umkehrwalze':
   for n, c in enumerate(self._alphabet):
    assert wiring[n] != c, '{}: Character {} reflected to itself'.format(self.__class__.__name__, c)
  
 def copy(self) -> Component:
  return copy.deepcopy(self)
 
 def nCC(self) -> int:
  ncc = 0
  for n, c in enumerate(self._wiring):
   ncc += (self._alphabet[n] != c)
  return ncc
  
 @property
 def name(self) -> str:
  """
  :getter: Returns the name of the componenrt  :setter: None
  """
  return self._name
 
 @property
 def alphabet(self) -> str:
  """
  :getter: Returns the alphabet of the engine
  :setter: None
  """
  return self._alphabet

 @property
 def numberOfPositions(self) -> int:
  """
  :getter: Returns the number of positions, usually len(alphabet)
  :setter: None
  """
  return self._numberOfPositions

 @property
 def wiring(self) -> str:
  """
  :getter: Returns the wiring
  :setter: Sets wiring and reverse wiring
  """
  return self._wiring

 @wiring.setter
 def wiring(self, newWiring : str) -> None:
  assert len(newWiring) == self._numberOfPositions,'{} {}: len(wiring) = {} != len(alphabet) = {}'.format(self.__class__.__name__, self._name, len(newWiring), self._numberOfPositions)
  _wiring = ''
  for n, c in enumerate(newWiring): 
   assert c in self.alphabet, '{}/wiring: Character {} not in {}'.format(self.__class__.__name__, c, self.alphabet)
   assert c not in _wiring, '{}/wiring: Duplicate character {} in wiring'.format(self.__class__.__name__, c)
   _wiring += c
  self._wiring = _wiring
  if '_rwiring' in vars(self):
   rwiringList = [' '] * self._numberOfPositions
   for n, c in enumerate(self._wiring):
    nr = self.alphabet.index(c)
    rwiringList[nr] = self.alphabet[n]
   self._rwiring = ''.join(rwiringList)
 
 @property
 def nextComponent(self) -> str:
  """
  :getter: Returns the next component in the chain
  :setter: Sets wiring and reverse wiring, if any 
  """
  return self._nextComponent
  
 @nextComponent.setter
 def nextComponent(self, component : Component) -> None:
  self._nextComponent = component

 @property
 def prevComponent(self) -> str:
  """
  :getter: Returns the previous component in the chain
  :setter: Sets wiring and reverse wiring
  """
  return self._prevComponent
  
 @prevComponent.setter
 def prevComponent(self, component : Component) -> None:
  if self.__class__.__name__ == 'Umkehrwalze':
   raise TypeError('{} is unidirectional'.format(self.__class__.__name__))
  self._prevComponent = component

 def encode(self, c : str, forward : bool, componentOnly : bool = False) -> str:
  """Runs forward or backward through a component

:param c: single character
:param forward: direction
:param componentOnly: do not go to next or previous component
:returns: encoded character (mapped back to the alphabet) 
  """
  assert len(c) == 1, '{}.encode({})/{}: len({}) != 1'.format(self.__class__.__name__, forward, self.name, c)
  assert c in self._alphabet, '{}.encode({})/{}: Character {} out of range'.format(self.__class__.__name__, forward, self.name, c)
  if '_ringstellung' in vars(self):
   state = self._alphabet.index(self._ringstellung)
  else:
   state = 0
  if forward:
   output = self._wiring[(self._alphabet.index(c) + state) % self._numberOfPositions]            
  else:
   output = self._alphabet[(self._alphabet.index(self._rwiring[self._alphabet.index(c)]) - state)  % self._numberOfPositions]     
  if self.notify is not None:
   self.notify('{}.encode({})/{}: {} => {}'.format(self.__class__.__name__, forward, self.name, c, output))
  if not componentOnly:
   if self._nextComponent and self.__class__ == Umkehrwalze:
    return self._nextComponent.encode(output, forward = False)
   elif self._nextComponent and forward:
    return self._nextComponent.encode(output, forward = True)
   elif self._prevComponent and not forward:
    return self._prevComponent.encode(output, forward = False)
  return output

 def chain(self, forward : bool = True) -> List[Tuple[Component, bool]]:
  """Creates a chain of visited components

:param forward: direction
:param _chain: current chain list (do not use)
:returns: (component, forward) list
  """
  chainList = list()
  lastComponent = self
  while lastComponent is not None:
   chainList.append((lastComponent, forward or lastComponent.__class__ == Umkehrwalze))
   if self.notify is not None:
    self.notify('{}.chain({})/{}'.format(self.__class__.__name__, forward, self.name))
   if lastComponent.__class__ == Umkehrwalze:
    lastComponent = lastComponent.nextComponent
    forward = False
   elif forward:
    lastComponent = lastComponent.nextComponent
   else:
    lastComponent = lastComponent.prevComponent
  return chainList
  
 def wiringToDict(self, connectedOnly : bool = False) -> Dict[str, str]:
  """Creates a dictionary starting from the wiring

:param connectedOnly: include only characters not connected to itself
:returns: dictionary of alphabet vs. wiring character
  """
  wiringDict = dict()
  for nc, c in enumerate(self.alphabet):
   if c != self.wiring[nc] or not connectedOnly:
    wiringDict[c] = self.wiring[nc]
  return wiringDict  

 def dictToWiring(self, wiringDict : Dict[str, str]) -> None:
  """Creates a dictionary starting from the wiring

:param wiringDict: dictionary of (at least connected) wirings
  """
  self.wiring = list(self.alphabet)
  for src, tgt in wiringDict.items():
   self.wiring[self.alphabet.index(src)] = tgt
  self.wiring = ''.join(self.wiring)  

 def __eq__(self, component : Component) -> bool:
  return self._name == component.name \
     and self._alphabet == component.alphabet \
     and self._wiring == component.wiring 

 def __repr__(self) -> str:
  return 'class: {}\nname: {}\nalphabet: {}\nwiring: {}'.format(
              self.__class__.__name__, self._name, self._alphabet, self._wiring)

class Steckerbrett(Umkehrwalze):
 """Represents a Steckerbrett (plugboard)

.. csv-table:: Wheel Controls
   :header: "Item", "Value"
   :widths: 30, 30
   
   *encryption*, forward and backward
   *ring*, none
   *notches*, none 

:param name: name of the Steckerbrett
:param wiring: wiring of the Steckerbrett, positions related to the *alphabet*
:param alphabet: unencoded alphabet to be used
:param notify: notification function (e.g. print)
 """
 def __init__(self, 
  name  : str = '', 
  wiring : Union[str, Dict[str, str]] = '', 
  alphabet  : str = stdAlphabet, 
  notify : Optional[Callable[[str], None]] = None) -> None:
  self._rwiring = ''
  super(Steckerbrett, self).__init__(name, wiring, alphabet, notify)
  if self.__class__ == Steckerbrett:
   if name == 'Mark 1':
    for group, groupWiring in zip([alphabet[:len(alphabet)//2], alphabet[len(alphabet)//2:]], [wiring[:len(alphabet)//2], wiring[len(alphabet)//2:]]):
     for ec in groupWiring:
      assert ec in group, '{} {}: Character {} not in {}'.format(self.__class__.__name__, self._name, ec, group)
   elif name == 'Mark 3':
    for c, ec in zip(alphabet, wiring):
     ecIndex = alphabet.index(ec)
     assert c == ec or c == wiring[ecIndex], '{} {}: Invalid connection {} -> {}'.format(self.__class__.__name__, self._name, c, ec)
 
 @staticmethod
 def Mark_1(
  replacement : bool = True, 
  alphabet : str = stdAlphabet, 
  notify : Optional[Callable[[str], None]] = None) -> Steckerbrett:
  """
  Random Initialization of a Mark 1 Steckerbrett (`crytomuseumSB`_)
  The contacts are splitted into 2 separated groups (sockets 1-13 and sockets 14-26)
  
.. _crytomuseumSB: https://www.cryptomuseum.com/crypto/enigma/i/sb.htm
  """
  wiring = str()
  for group in [alphabet[:len(alphabet)//2], alphabet[len(alphabet)//2:]]:
   while True: 
    actWiring = ''.join(random.sample(group, len(group)))
    hasReplacement = False
    if not replacement:
     for c, ec in zip(group, actWiring):
      if c == ec:
       hasReplacement = False
    if not hasReplacement:
     wiring += actWiring
     break
  return Steckerbrett('Mark 1', wiring, alphabet, notify)
     
 @staticmethod
 def Mark_2(
  replacement : bool = True, 
  alphabet : str = stdAlphabet, 
  notify : Optional[Callable[[str], None]] = None) -> Steckerbrett:
  """
  Random Initialization of a Mark 2 Steckerbrett (`crytomuseumSB`_)
  The 52 contacts are connected in unlimited fashion
  
.. _crytomuseumSB: https://www.cryptomuseum.com/crypto/enigma/i/sb.htm
  """
  wiring = str()
  while True: 
   actWiring = ''.join(random.sample(alphabet, len(alphabet)))
   hasReplacement = False
   if not replacement:
    for c, ec in zip(alphabet, actWiring):
     if c == ec:
      hasReplacement = False
   if not hasReplacement:
    wiring += actWiring
    break
  return Steckerbrett('Mark 2', wiring, alphabet, notify)

 @staticmethod
 def Mark_3(
  nConnections : int = 10, 
  alphabet : str = stdAlphabet, 
  notify : Optional[Callable[[str], None]] = None) -> Steckerbrett:
  """
  Random Initialization of a Mark 3 Steckerbrett (`crytomuseumSB`_)
  The 26 contacts are cross-connected, i.e. MU is combined with UM
  
.. _crytomuseumSB: https://www.cryptomuseum.com/crypto/enigma/i/sb.htm
  """
  assert nConnections >= 0 and nConnections <= len(alphabet) // 2, '{}: 0 < nConnections = {} < {} required'.format(Steckerbrett.__class__.__name__, nConnections, len(alphabet) // 2)
  sources = ''.join(random.sample(alphabet, nConnections))
  targetAlphabet = str()
  for c in alphabet:
   if c not in sources:
    targetAlphabet += c
  targets = ''.join(random.sample(targetAlphabet, nConnections))
  wiring = list(alphabet)
  for src, tgt in zip(sources, targets):
   srcIndex = alphabet.index(src)
   tgtIndex = alphabet.index(tgt)
   wiring[srcIndex] = tgt
   wiring[tgtIndex] = src
  return Steckerbrett('Mark 3', ''.join(wiring), alphabet, notify)
  
 def isMark1(self) -> bool:
  """
  Checks, if the steckerbrett is of type Mark1

:returns: indicator
  """
  halfAlphabet2 = self.alphabet[len(self.alphabet)//2:]
  for c in self.alphabet[:len(self.alphabet)//2]:
   if c not in halfAlphabet2:
    return False
  return True

 def isMark3(self) -> bool:
  """
  Checks, if the steckerbrett is of type Mark3
  
:returns: indicator
  """
  return self.wiring == self.rwiring
  
 def nConnections(self) -> int:
  """
  Counts the number of connections of a steckerbrett
  """
  nConnections2 = 0
  for n, c in enumerate(self.alphabet):
   if self.wiring[n] != c:
    nConnections2 += 1
  return nConnections2 // 2

 @staticmethod
 def _replaceChar(actStr : str, index : str, newStr : str) -> str:
  return actStr[:index] + newStr + actStr[index+1:]

 def clearMark3Setting(self, c : str) -> None:
  """
  Clear one setting of a Steckerbrett (`crytomuseumSB`_)
  """
  assert len(c) == 1, '{}.clearMark3Setting: {} is not a single character'.format(self.__class__.__name__, c)
  assert c in self.alphabet, '{}.clearMark3Setting: {} is not in alphabet'.format(self.__class__.__name__, c)
  srcIndex = self.alphabet.index(c)
  tgtIndex = self.alphabet.index(self.wiring[srcIndex])
  if srcIndex == tgtIndex:
   return
  assert c == self.wiring[tgtIndex], '{}.clearMark3Setting: {} is not in Mark 3 steckerbrett'.format(self.__class__.__name__, self.name)
  wiring = self.wiring
  wiring = self._replaceChar(wiring, srcIndex, self.alphabet[srcIndex])
  wiring = self._replaceChar(wiring, tgtIndex, self.alphabet[tgtIndex])
  self.wiring = wiring

 def addMark3Setting(self, src : str, tgt : str) -> None:
  """
  Add one setting to a Steckerbrett (`crytomuseumSB`_)
  """
  assert len(src) == 1 and len(tgt) == 1, '{}.addMark3Setting: Either src = {} and tgt = {} are not single characters'.format(
             self.__class__.__name__, src, tgt)
  assert src in self.alphabet and tgt in self.alphabet, '{}.addMark3Setting: Either src = {} and tgt = {} is not in alphabet'.format(
             self.__class__.__name__, src, tgt)
  if src == tgt:
   self.clearMark3Setting(src)
  srcIndex = self.alphabet.index(src)
  tgtIndex = self.alphabet.index(tgt)
  if src == self.wiring[tgtIndex] and tgt == self.wiring[srcIndex]:
   return
  assert src == self.wiring[srcIndex], '{}.addMark3Setting: src = {} is already connected'.format(self.__class__.__name__, src)
  assert tgt == self.wiring[tgtIndex], '{}.addMark3Setting: tgt = {} is already connected'.format(self.__class__.__name__, tgt)
  wiring = self.wiring
  wiring = self._replaceChar(wiring, srcIndex, tgt)
  wiring = self._replaceChar(wiring, tgtIndex, src)
  self.wiring = wiring

 def exchangeMark3Setting(self, c1 : str, c2 : str) -> None:
  """
  Exchanges 2 settings of a Steckerbrett (`crytomuseumSB`_)
  """
  assert len(c1) == 1 and len(c2) == 1, '{}.exchangeMark3Setting: Either src = {} and tgt = {} are not single characters'.format(
             self.__class__.__name__, c1, c2)
  assert c1 in self.alphabet and c2 in self.alphabet, '{}.exchangeMark3Setting: Either src = {} and tgt = {} is not in alphabet'.format(
             self.__class__.__name__, c1, c2)
  if c1 == c2:
   return
  c1Index = self.wiring.index(c1)
  assert c1 != self.alphabet[c1Index], '{}.exchangeMark3Setting: c1 = {} is unconnected'.format(self.__class__.__name__, c1)
  c2Index = self.wiring.index(c2)
  assert c2 != self.alphabet[c2Index], '{}.exchangeMark3Setting: c2 = {} is unconnected'.format(self.__class__.__name__, c2)
  wiring = self.wiring
  wiring = self._replaceChar(wiring, c1Index, c2)
  wiring = self._replaceChar(wiring, c2Index, c1)
  self.wiring = wiring

 @property
 def rwiring(self) -> str:
  """
  :getter: Returns the reverse wiring
  """
  return self._rwiring

 def __repr__(self) -> str:
  return super(Steckerbrett, self).__repr__() + '\nreverse wiring: {}'.format(self._rwiring)

class Zusatzwalze(Steckerbrett) : 
 """Represents an Zusatzwalze (ZW == extra rotor)

.. csv-table:: Wheel Controls
   :header: "Item", "Value"
   :widths: 30, 30
   
   *encryption*, forward and backward
   *ring*, yes
   *notches*, none 

:param name: name of the Zusatzwalze
:param wiring: wiring of the Zusatzwalze, positions related to the *alphabet*
:param alphabet: unencoded alphabet to be used
:param notify: notification function (e.g. print)
 """
 def __init__(self, 
  name  : str = '', 
  wiring : Union[str, Dict[str, str]] = '', 
  ringstellung : Optional[str] = None, 
  alphabet  : str = stdAlphabet, 
  notify : Optional[Callable[[str], None]] = None) -> None:
  """
  Initialization
  """
  super(Zusatzwalze, self).__init__(name, wiring, alphabet, notify)
  self._prevComponent : Optional[Component] = None
  if ringstellung is None:
   self._ringstellung = self._alphabet[0]
  else:
   assert len(ringstellung) == 1
   assert ringstellung in self._alphabet
   self._ringstellung = ringstellung

 @property
 def ringstellung(self) -> str:
  """
  :getter: Returns the ringstellung
  :setter: Sets the ringstellung
  """
  return self._ringstellung
  
 @ringstellung.setter
 def ringstellung(self, ringstellung : str) -> None:
   assert len(ringstellung) == 1, '{}.ringstellung {}: 1 character expected, {} got'.format(self.__class__.__name__, self._name, ringstellung)
   assert ringstellung in self._alphabet
   self._ringstellung = ringstellung

 def __eq__(self, component : Component) -> bool:
  return super(Zusatzwalze, self).__eq__(component) and self._ringstellung == component._ringstellung

 def __repr__(self) -> str:
  return super(Zusatzwalze, self).__repr__() + '\nringstellung: {}'.format(self._ringstellung)

class Walze(Zusatzwalze):
 """Represents a Walze (rotor)

.. csv-table:: Wheel Controls
   :header: "Item", "Value"
   :widths: 30, 30
   
   *encryption*, forward and backward
   *ring*, yes
   *notches*, yes 

:param name: name of the walze
:param wiring: wiring of the walze, positions related to the *alphabet*
:param ringstellung: visible character at the beginning, position related to the *alphabet*
:param notches: visible characters for the notches, positions related to the *alphabet*
:param alphabet: unencoded alphabet to be used
:param notify: notification function (e.g. print)

 """
 def __init__(self, 
  name  : str = '', 
  wiring : Union[str, Dict[str, str]] = '', 
  ringstellung : Optional[str] = None, 
  notches : str  = '', 
  alphabet : str = stdAlphabet, 
  notify : Optional[Callable[[str], None]] = None) -> None:
  """
  Initialization
  """
  super(Walze, self).__init__(name, wiring, ringstellung, alphabet, notify)
  assert notches, '{} {}: At least 1 notch required'.format(self.__class__.__name__, self._name)
  self._notches = ''
  for notch in notches:
   assert notch in self._alphabet, '{} {}: Notch {} out of range'.format(self.__class__.__name__, self._name, notch)
   assert notch not in self._notches, '{} {}: Duplicate notch {}'.format(self.__class__.__name__, self._name, notch)
   self._notches += notch
   
 def stepN(self, n : int)-> None:
  """
  Steps the Walze n times
  """
  assert n > 0, '{} {}: n == {} <= 0'.format(self.__class__.__name__, self._name, n)
  for _ in range(n):
   self.step()

 def step(self)-> None:
  """
  Steps the Walze and inititiates of the next Walze if available
  
   III	II	I	<-- wheel order
   A	D	O
   A	D	P
   A	D	Q
   A	E	R	<-- 1st step of middle wheel
   B	F	S	<-- 2nd step of middle wheel
   B	F	T
   B	F	U  
  """
  notchFound = self._ringstellung in self._notches
  if notchFound:
   if self._nextComponent and self._nextComponent.__class__.__name__ == 'Walze':
    self._nextComponent.step()
  self._ringstellung = self._alphabet[(self._alphabet.index(self._ringstellung) + 1) % self._numberOfPositions]
  if self.notify:
   self.notify('{}.step {}: {}, notch = {}'.format(self.__class__.__name__, self.name, self._ringstellung, notchFound))

 def __eq__(self, component : Component) -> bool:
  return super(Walze, self).__eq__(component) and self._notches == component._notches

 def __repr__(self) -> str:
  return super(Walze, self).__repr__() + '\nnotches: {}'.format(self._notches)

 @property
 def notches(self) -> str:
  """
  :getter: Returns the notches
  :setter: None
  """
  return self._notches



# from typing import Dict, Tuple, Any
import pytest

import re
import copy
import MzEnigma

def test_steckerbrett(pytestconfig):
 if pytestconfig.getoption('wiring'):
  for name, cls in vars(MzEnigma).items():
   if cls.__class__ == MzEnigma.Steckerbrett:
    print('\n -{}'.format(cls))
    
def test_ringstellung(pytestconfig):
 if pytestconfig.getoption('wiring'):
  print('\n--- test_ringstellung ---')
  for name, cls in vars(MzEnigma).items():
   if isinstance(cls, MzEnigma.Walze):
    encoded = str()
    for c in cls.alphabet:
     cls.ringstellung = c
     encoded += cls.encode(cls.alphabet[0], True)
    assert encoded == cls.wiring, 'Ringstellung of component {} failed'.format(cls.name)
    print('- {} completed'.format(cls.name))

def test_wiring(pytestconfig):
 if pytestconfig.getoption('wiring'):
  print('\n--- test_wiring ---')
  for name, cls in vars(MzEnigma).items():
   if isinstance(cls, MzEnigma.Steckerbrett):
    cls.notify = pytest.helpers.notify(pytestconfig)
    alphabet = cls.alphabet
    pos = 17
    if '_ringstellung' in vars(cls):
     cls.ringstellung = alphabet[pos]
    encoded = str()
    for c in alphabet:
     encoded += cls.encode(c, True, componentOnly = True)
    if '_ringstellung' in vars(cls):
     expectedWiring = cls.wiring[pos:] + cls.wiring[:pos]
    else:
     expectedWiring = cls.wiring
    assert encoded == expectedWiring, 'Reverse wiring of component {} failed, expected = {}, got = {}'.format(cls.name, encoded, expectedWiring)
    got = str()
    for n, c in enumerate(alphabet):
     got += cls.encode(encoded[n], False)
    assert alphabet == got, 'Reverse wiring of component {} failed, expected = {}, got = {}'.format(cls.name, encoded,  got)
    print('- {} completed'.format(cls.name))
 
def test_step(pytestconfig):
 if pytestconfig.getoption('wiring'):
  print('\n--- test_step ---')
  for name, cls1 in vars(MzEnigma).items():
   if isinstance(cls1, MzEnigma.Walze):
    cls1.ringstellung = cls1.alphabet[0]
    cls2 = copy.deepcopy(cls1)
    cls1.nextComponent = cls2
    ringstellung1 = cls1.alphabet.index(cls1.ringstellung)
    ringstellung2 = ringstellung1
    for c in cls1.alphabet:
     ringstellung1 = (ringstellung1 + 1) % len(cls1.alphabet)
     ringstellung2 += int(cls1.ringstellung in cls1.notches)
     cls1.step()
     assert cls1.alphabet[ringstellung1] == cls1.ringstellung and cls1.alphabet[ringstellung2] == cls2.ringstellung, 'Stepping of component {} failed'.format(cls1.name)
    print('- {} completed'.format(cls1.name))

if __name__ == "__main__":
 import sys
 pattern = 'I_.+'
 for name, cls in vars(MzEnigma).items():
  if isinstance(cls, MzEnigma.Umkehrwalze) and re.fullmatch(pattern, name):
   print('\n{}'.format(cls))
 pytest.main([__file__])  
 print(MzEnigma.I_A)




"""
.. _CrypTool: https://www.cryptool.org/en/documentation/ctbook/
.. _Decrypted Secrets: https://link.springer.com/book/10.1007/978-3-540-48121-8

The cryptoanalysis methods implemented here include:

.. csv-table:: 
   :header: "Author", "Requirement", "See ..."
   :widths: 15, 30, 30
   
   *Rejewski*, encoded Spruchschluessel, `Decrypted Secrets`_
   *Turing*, unencoded crib, `Decrypted Secrets`_
   *Gillogly*, language statistics, `CrypTool`_
   
None of these methods always succeed. As a rule of thumb, the
Turing and Gillogly attacks require at least 30 characters
to get sufficient analysis results.
"""

from __future__ import annotations
from typing import Optional, Callable, Dict, List, Tuple

import copy
import random
import itertools
import pickle
import os.path

import networkx

import MzEnigma

class TagesschluesselRange(object):
 """Represents a range of Tagesschluessel and analysis methods working on them
 
:param enigma: model of the Enigma (required)
:param umkehrwalzenList: 'Umkehrwalze' (if undefined, randomly selected)
:param walzenList: List of type 'Walze' (if undefined, randomly selecte(if undefined, randomly selected)
:param tagesWalzenStellungenList: List of Tageswalzenstellungen (if undefined, all) 
:param spruchScoring: 'SpruchScoring' (required by Gillolgy attack)
:param zusatzwalzenList: 'Zusatzwalze' (if undefined, randomly selected)
:param blank: replacement character for a blank
:param notify: notification function (e.g. print)
  """
 def __init__(self,
  enigma : Optional[MzEnigma.Enigma] = None,  
  umkehrwalzenList : Optional[List[MzEnigma.Umkehrwalze]] = None, 
  walzenList : Optional[List[MzEnigma.Walze]] = None,
  tagesWalzenStellungenList : Optional[List[str]] = None,
  zusatzwalzenList : Optional[List[MzEnigma.Zusatzwalze]] = None, 
  spruchScoring : Optional[MzEnigma.SpruchScoring] = None, 
  blank : str = '',  
  notify : Optional[Callable[[str], None]] = None) -> None:
  self.notify = notify
  assert enigma, '{}: Enigma required'.format(self.__class__.__name__)
  self.enigma = enigma
  self.blank = blank
  alphabet = self.enigma.alphabet
  if enigma.steckerbrett:
   self.steckerbrett = copy.deepcopy(enigma.steckerbrett)
   self.steckerbrett.notify = self.notify
  else:
   self.steckerbrett = None
  assert walzenList is not None and len(walzenList) > 0, '{}: WalzenRange is required'.format(self.__class__.__name__)
  if walzenList is None or len(walzenList) == 0:
   walzenList = itertools.permutations(self.enigma.walzenList, self.enigma.numberOfWalzen)
  self.walzenList = list()
  for _walzen in walzenList:
   walzen = list()
   for _walze in _walzen:
    walze = copy.deepcopy(_walze)
    walze.notify = self.notify
    walzen.append(_walze)
   self.walzenList.append(walzen)
  if tagesWalzenStellungenList:
   for tagesWalzenStellungen in tagesWalzenStellungenList:
    assert len(tagesWalzenStellungen) == enigma.numberOfWalzen, '{}: Walzenstellung, number of characters must match the number of walzen'.format(self.__class__.__name__)
    assert all(v in alphabet for v in tagesWalzenStellungen), '{}: Walzenstellung, all characters must be in alphabet'.format(self.__class__.__name__)
  else:
   tagesWalzenStellungenList = list(itertools.product(self.enigma.alphabet, repeat = self.enigma.numberOfWalzen))
  self.tagesWalzenStellungenList = copy.deepcopy(tagesWalzenStellungenList )
  self.zusatzwalzenList = list()
  if enigma.zusatzwalzen is not None:
   assert zusatzwalzenList is not None, '{}: zusatzwalzenList is required'.format(self.__class__.__name__)
   for _walze in zusatzwalzenList:
    walze = copy.deepcopy(_walze)
    walze.notify = self.notify
    self.zusatzwalzenList.append(walze)
  assert umkehrwalzenList is not None or len(self.enigma.umkehrwalzen) == 1, '{}: umkehrwalzenList is required'.format(self.__class__.__name__)
  if umkehrwalzenList is None:
   umkehrwalzenList = [ self.enigma.umkehrwalzen ]
  self.umkehrwalzenList = list()
  for _walze in umkehrwalzenList:
   walze = copy.deepcopy(_walze)
   walze.notify = self.notify
   self.umkehrwalzenList.append(walze)
  self.spruchScoring = spruchScoring 
 
 @classmethod
 def rejewskiAttack(cls, catalog : List[Tuple[MzEnigma.Tagesschluessel, List[List[str]]]], encodedSpruchSchluessel : str) -> List[Tuple[MzEnigma.Tagesschluessel, str]]:
  """Find candidate values for the unencoded SpruchSchluessel from a catalog

:param catalog: catalog of encoded spruchschluessels (required)
:param encodedSpruchSchluessel: encoded spruchschluessels to be found (required)
:returns: List[Tuple[Tagesschluessel, unencodedSpruchSchluessel]]
  """
  cls.checkRejewskiCatalog(catalog)
  firstKey, firstItem = catalog[0]
  alphabet = firstKey.alphabet

  encodedPosKeys = list()
  for first, second in zip(encodedSpruchSchluessel[:len(firstItem)], encodedSpruchSchluessel[len(firstItem):]):
   encodedPosKeys.append(first+second)
   
  validCandidates = list()
  for tagesschluessel, listOfDoublets in catalog:
   unencodedPosKeys = list()
   for pos, doublets in enumerate(listOfDoublets):
    if encodedPosKeys[pos] not in doublets:
     break
    unencodedPosKeys.append(alphabet[doublets.index(encodedPosKeys[pos])])
   if len(unencodedPosKeys) == len(listOfDoublets):
    validCandidates.append((tagesschluessel,''.join(unencodedPosKeys)))
   
  return validCandidates

 def createRejewskiCatalog(self, pickleFile : Optional[str] = None) -> List[Tuple[MzEnigma.Tagesschluessel, List[List[str]]]]:
  """Build up a catalog of encoded spruchschluessels. Works only without a steckerbrett. 

:param pickleFile: pickle file to be created (optional)
:returns: List[Tuple[Tagesschluessel, List[List[Doublet]]]
  """
  assert self.enigma.steckerbrett is None, '{}.createRejewskiCatalog: engines with steckerbrett are not supported'.format(self.__class__.__name__)
  if pickleFile is not None:
   pickleFile = os.path.normpath(pickleFile)
   assert not os.path.exists(pickleFile), '{}.createRejewskiCatalog: pickle file {} is already exising'.format(self.__class__.__name__, pickleFile)
  gTagesschluessel = MzEnigma.Tagesschluessel(enigma = self.enigma, notify = None)
  rejewskiList = list()
  for umkehrwalze in self.umkehrwalzenList:
   if self.notify:
    self.notify('{}.createRejewskiCatalog: - examining umkehrwalze = {}'.format(self.__class__.__name__, umkehrwalze.name))
   for walzen in self.walzenList:
    if self.notify:
     nameList = list()
     for walze in walzen:
      nameList.append(walze.name)
     self.notify('{}.createRejewskiCatalog:  - examining walzen = {}'.format(self.__class__.__name__, nameList))
    for tagesWalzenStellungenID, tagesWalzenStellungenTuple in enumerate(self.tagesWalzenStellungenList):
     tagesWalzenStellungen = ''.join(tagesWalzenStellungenTuple)
     if self.notify:
      self.notify('{}.createRejewskiCatalog:   + examining tagesWalzenStellungen = {} ({} of {})'.format(self.__class__.__name__, tagesWalzenStellungen, tagesWalzenStellungenID,  len(self.tagesWalzenStellungenList)))
     if len(self.zusatzwalzenList) > 0:
      for zusatzwalze in self.zusatzwalzenList:
       if self.notify:
        self.notify('{}.createRejewskiCatalog:  - examining zusatzwalze = {}'.format(self.__class__.__name__, zusatzwalze.name))
       tagesschluessel = MzEnigma.Tagesschluessel.changeWalzen(
                                  gTagesschluessel, walzen = walzen, tagesWalzenStellungen = tagesWalzenStellungen, umkehrwalze = umkehrwalze, zusatzwalze = zusatzwalze)
       rejewskiList.append((tagesschluessel, tagesschluessel.findDoublets(first = 0, second = self.enigma.numberOfWalzen, all = True)))
     else:
      tagesschluessel = MzEnigma.Tagesschluessel.changeWalzen(
                                 gTagesschluessel, walzen = walzen, tagesWalzenStellungen = tagesWalzenStellungen, umkehrwalze = umkehrwalze)
      rejewskiList.append((tagesschluessel, tagesschluessel.findDoublets(first = 0, second = self.enigma.numberOfWalzen, all = True)))
  if pickleFile is not None:
   pickle.dump( rejewskiList, open( pickleFile, "wb" ) )
  return rejewskiList
  
 @classmethod
 def checkRejewskiCatalog(cls, catalog : List[Tuple[MzEnigma.Tagesschluessel, List[List[str]]]], expectedSpruchschluesselLength : Optional[int] = None) -> None:
  """check a catalog of encoded spruchschluessels  

:param expectedSpruchschluesselLength: number of characters in unencoded spruchschluessel (optional)
  """
  assert len(catalog) > 0, '{}.checkRejewskiCatalog: empty catalog detected'.format(cls.__class__.__name__)  
  firstKey, firstItem = catalog[0]
  assert isinstance(firstKey, MzEnigma.Tagesschluessel), '{}.checkRejewskiCatalog: improper catalog key, MzEnigma.Tagesschluessel espected'.format(cls.__class__.__name__)
  assert isinstance(firstItem, list), '{}.checkRejewskiCatalogk: improper catalog item, list espected'.format(cls.__class__.__name__)  
  alphabet = firstKey.alphabet
  if expectedSpruchschluesselLength is not None:
   assert len(firstItem) == expectedSpruchschluesselLength, '{}.checkRejewskiCatalog: number of positions in catalog item ({}) != number of positions in spruchschluessel ({})'.format(cls.__class__.__name__, len(firstItem[0]), expectedSpruchschluesselLength)  
  assert len(firstItem[0]) == len(alphabet), '{}.checkRejewskiCatalog: number of positions in catalog item.item({}) != number of characters ({})'.format(cls.__class__.__name__, len(firstItem), len(alphabet))  
  assert isinstance(firstItem[0][0], str), '{}.checkRejewskiCatalog: improper catalog item.item, str espected'.format(cls.__class__.__name__)  

 @classmethod
 def loadRejewskiCatalog(cls, pickleFile : str) -> List[Tuple[MzEnigma.Tagesschluessel, List[List[str]]]]:
  """Build up a catalog of encoded spruchschluessels  

:returns: List[Tuple[Tagesschluessel, List[List[Doublet]]]
  """
  pickleFile = os.path.normpath(pickleFile)
  assert os.path.isfile(pickleFile), '{}.loadRejewskiCatalog: pickle file {} is not or not a file'.format(cls.__class__.__name__, pickleFile)
  rejewskiList = pickle.load( open( pickleFile, "rb" ) )
  cls.checkRejewskiCatalog(rejewskiList)
  return rejewskiList

 def turingAttack(self, encodedSpruch : str = '', crib : str = '', startingPosition : int = 0) -> List[Tuple[MzEnigma.Tagesschluessel, List[Tuple[str, str]]]]:
  """Turing attack to derive candidate settings for Umkehrwalze, Walzen, Zusatzwalze, Tageswalzenstellungen and several settings of the Steckerbrett
An engine with Steckerbrett is required.
The rate of correct Tagesschluessels increases with increasing length of the crib.

:param encodedSpruch: encoded message to be attacked (required)
:param crib: unencoded crib to be found in the message (required) 
:param startingPosition: starting position of the crib in the message
:return: List[Tuple[Tagesschluessel, List[Tuple[Walzenstellungen, Steckerbrett.wiring]]]]
  """
  def connectBus(id : int, v : str) -> bool:
   '''Heart of the turing bomb - the bus is restricted to the ONE active line with a character
   
   :param id: bus to be used
   :param v: input value
   :return: boolean indicating success
   '''
   nonlocal graph, ecList, buses
   if buses[id] is None:
    buses[id] = v
    vID = self.enigma.alphabet.index(v)
    for tgtID in graph.neighbors(id):
     for _, posDict in graph[id][tgtID].items():
      pos = posDict['pos']
      tgtV = ecList[vID][pos]
      if not connectBus(tgtID, tgtV):
       return False
    return True
   else:
    return buses[id] == v
   
  def getValidWalzenStellungen(tagesschluessel : MzEnigma.Tagesschluessel) -> List[Tuple[str, List[Dict[str, str]]]]: 
   nonlocal ecList, buses, firstBus, spruchLen, graph
   if self.notify:
    self.notify('{}.turingAttack: \n{}'.format(self.__class__.__name__, tagesschluessel))
   oldTagesWalzenStellungen = tagesschluessel.tagesWalzenStellungen
   candidateSteckerbrettList = list()
   for tagesWalzenStellungenID, tagesWalzenStellungenTuple in enumerate(self.tagesWalzenStellungenList):
    knownWirings = list()
    tagesschluessel.tagesWalzenStellungen = ''.join(tagesWalzenStellungenTuple)
    if self.notify:
     self.notify('{}.turingAttack:  - examining tagesWalzenStellungen = {} ({} of {})'.format(self.__class__.__name__, tagesschluessel.tagesWalzenStellungen, tagesWalzenStellungenID,  len(self.tagesWalzenStellungenList)))
    ecList = list()
    for nc, c in enumerate(self.enigma.alphabet):
     ecList.append(list())
     for cs in tagesschluessel.encode(spruchLen * c):
      ecList[-1].append(cs)
    
    for v in self.enigma.alphabet:
     if self.notify:
       self.notify('{}.turingAttack:  + examining "{}"'.format(self.__class__.__name__, v))
     buses = len(self.enigma.alphabet) * [ None ]
     if connectBus(firstBus, v):
      knownWiring = dict()
      for n, id in enumerate(buses):
       if id is not None:
        knownWiring[self.enigma.alphabet[n]] = id
      keys = list(knownWiring.keys())
      for key in keys:
       value = knownWiring[key]
       if value not in keys:
        knownWiring[value] = key
      knownWirings.append(knownWiring)
      if self.notify:
       self.notify('{}.turingAttack:   wiring {} found'.format(self.__class__.__name__, knownWiring))
    candidateSteckerbrettList.append((tagesschluessel.tagesWalzenStellungen, knownWirings))
   tagesschluessel.tagesWalzenStellungen = oldTagesWalzenStellungen
   return candidateSteckerbrettList

  assert self.enigma.steckerbrett is not None, '{}.turingAttack: engines without steckerbrett are not supported'.format(self.__class__.__name__)
  posList = MzEnigma.Enigma.validCribPositions(encodedSpruch, crib)
  assert startingPosition in posList, '{}.turingAttack: {} is not a valid starting position, use in {}'.format(self.__class__.__name__, startingPosition, posList)
  encodedSpruch = encodedSpruch[startingPosition:]
  ecList = list()
  buses = list()
  
  # We need to use a MultiGraph, since the menu may contain duplicate edges (with different positions) 
  graph = networkx.MultiGraph()

  for pos, cc in enumerate(crib):
   icc = self.enigma.alphabet.index(cc)
   iec = self.enigma.alphabet.index(encodedSpruch[pos])
   graph.add_edge(icc, iec, pos = pos)
  firstBus = sorted(graph.adjacency(), key = lambda item: len(item[1]), reverse = True)[0][0]
 
  spruchLen = len(crib)
  encodedSpruch = encodedSpruch[:spruchLen]
  
  gTagesschluessel = MzEnigma.Tagesschluessel(enigma = self.enigma, 
                                                                 steckerbrett = copy.deepcopy(MzEnigma.UnconnectedSteckerbrett), 
                                                                 tagesWalzenStellungen = self.enigma.numberOfWalzen * 'A', 
                                                                 blank = self.blank, notify = None)
  validCandidates = list()
  for umkehrwalze in self.umkehrwalzenList:
   if self.notify:
    self.notify('{}.turingAttack: - examining umkehrwalze = {}'.format(self.__class__.__name__, umkehrwalze.name))
   for walzen in self.walzenList:
    if self.notify:
     nameList = list()
     for walze in walzen:
      nameList.append(walze.name)
     self.notify('{}.turingAttack:  - examining walzen = {}'.format(self.__class__.__name__, nameList))
    if len(self.zusatzwalzenList) > 0:
     for zusatzwalze in self.zusatzwalzenList:
      if self.notify:
       self.notify('{}.turingAttack:  - examining zusatzwalze = {}'.format(self.__class__.__name__, zusatzwalze.name))
      tagesschluessel = MzEnigma.Tagesschluessel.changeWalzen(
                                 gTagesschluessel, walzen = walzen, tagesWalzenStellungen = gTagesschluessel.tagesWalzenStellungen, 
                                 umkehrwalze = umkehrwalze, zusatzwalze = zusatzwalze)
      candidateSteckerbrettList = getValidWalzenStellungen(tagesschluessel)
      if len(candidateSteckerbrettList) > 0:
       validCandidates.append((copy.deepcopy(tagesschluessel), candidateSteckerbrettList))
    else:
     tagesschluessel = MzEnigma.Tagesschluessel.changeWalzen( 
                                gTagesschluessel, walzen = walzen, tagesWalzenStellungen = gTagesschluessel.tagesWalzenStellungen, umkehrwalze = umkehrwalze)
     candidateSteckerbrettList = getValidWalzenStellungen(tagesschluessel)
     if len(candidateSteckerbrettList) > 0:
      validCandidates.append((copy.deepcopy(tagesschluessel), candidateSteckerbrettList))
  return validCandidates
  
 def gilloglyAttackPhase1(self, encodedSpruch : str) -> MzEnigma.Tagesschluessel:
  """Brute force attack to derive settings for Umkehrwalze, Walzen, Zusatzwalze, and Tageswalzenstellungen.
Uses the index of coincidence for scoring.
The rate of correct Tagesschluessels increases with increasing length of the encodedSpruch.
 
:param encodedSpruch: message to be attacked (required)
:return: Tagesschlüssel 
  """
  def doScoring():
   nonlocal encodedSpruch, bestScore, tagesschluessel, bestTagesschluessel, lastDecryptedSpruch
   decryptedSpruch = tagesschluessel.decode(encodedSpruch)
   actScore = MzEnigma.SpruchScoring.indexOfCoincidence(decryptedSpruch)
   # decryptedSpruchMD5 = hashlib.md5(str(decryptedSpruch).encode("utf-8")).hexdigest()
   if self.notify:
    self.notify('{}.gilloglyAttackPhase1: + examining tagesWalzenStellungen = {}'.format(self.__class__.__name__, tagesschluessel.tagesWalzenStellungen))
    self.notify('{} score = {:.3f} (best: {:.3f})'.format(60*' ', actScore, bestScore))
   lastDecryptedSpruch = decryptedSpruch
   if actScore > bestScore:
    bestScore = actScore
    bestTagesschluessel = copy.deepcopy(tagesschluessel)
   
  lastDecryptedSpruch = ''
  gTagesschluessel = MzEnigma.Tagesschluessel(enigma = self.enigma, steckerbrett = self.steckerbrett, blank = self.blank, notify = None)
  bestTagesschluessel = None
  bestScore = 0
  for umkehrwalze in self.umkehrwalzenList:
   if self.notify:
    self.notify('{}.gilloglyAttackPhase1: - examining umkehrwalze = {}'.format(self.__class__.__name__, umkehrwalze.name))
   for walzen in self.walzenList:
    if self.notify:
     nameList = list()
     for walze in walzen:
      nameList.append(walze.name)
     self.notify('{}.gilloglyAttackPhase1:  - examining walzen = {}'.format(self.__class__.__name__, nameList))
    for tagesWalzenStellungenTuple in self.tagesWalzenStellungenList:
     tagesWalzenStellungen = ''.join(tagesWalzenStellungenTuple)
     if len(self.zusatzwalzenList) > 0:
      for zusatzwalze in self.zusatzwalzenList:
       if self.notify:
        self.notify('{}.gilloglyAttackPhase1:  - examining zusatzwalze = {}'.format(self.__class__.__name__, zusatzwalze.name))
       tagesschluessel = MzEnigma.Tagesschluessel.changeWalzen(
                                  gTagesschluessel, walzen = walzen, tagesWalzenStellungen = tagesWalzenStellungen, umkehrwalze = umkehrwalze, zusatzwalze = zusatzwalze)
       doScoring()
     else:
      tagesschluessel = MzEnigma.Tagesschluessel.changeWalzen( 
                                 gTagesschluessel, walzen = walzen, tagesWalzenStellungen = tagesWalzenStellungen, umkehrwalze = umkehrwalze)
      doScoring()
  if self.notify:
   self.notify('{}.gilloglyAttackPhase1: bestScore = {:.3f}\n{}'.format(self.__class__.__name__,  bestScore, bestTagesschluessel))
  return bestTagesschluessel

 def shotgunPhase2(self, 
   phase1Tagesschluessel : MzEnigma.Tagesschluessel, encodedSpruch : str = '', noImprovement : int = 10) -> MzEnigma.Tagesschluessel:
  """Shotgun hill climbing or simulated annealing attack to derive settings for Steckerbrett
Uses the 'newNgramScore' for scoring.
 
:param phase1Tagesschluessel: tagesschluessel with correct settings for Umkehrwalze, Walzen, Zusatzwalze, and Tageswalzenstellungen(required)
:param encodedSpruch: encrypted message to be attacked (required)
:param noImprovement: stop condition after *noImprovement* number of attempts in a cycle
:return: Tagesschlüssel
  """
  def fillWired(tagesschluessel):
   wired = set()
   unwired = set()
   for n, c in enumerate(tagesschluessel.alphabet):
    if tagesschluessel.steckerbrett.wiring[n] != c:
     wired.add(c)
    else:
     unwired.add(c)
   return wired, unwired

  assert self.spruchScoring is not None, '{}.shotgunPhase2: self.spruchScoring is required'.format(self.__class__.__name__)
  assert phase1Tagesschluessel.steckerbrett is not None, '{}.gilloglyAttackPhase2: steckerbrett is required'.format(self.__class__.__name__)
  wired,  unwired = fillWired(phase1Tagesschluessel)
  if len(wired) <= 2:
   return phase1Tagesschluessel
  assert len(unwired) > 2, '{}.shotgunPhase2: steckerbrett with > 0 unconnected pins required'.format(self.__class__.__name__)
  tagesschluessel = copy.deepcopy(phase1Tagesschluessel)

  notImproved = 0
  bestScore = 0
  while notImproved < noImprovement:
   actScore = self.spruchScoring.newNgramScore(tagesschluessel.decode(encodedSpruch), bestScore, tagesschluessel.alphabet)
   if actScore >= bestScore:
    if actScore > bestScore:
     notImproved = -1
    bestScore = actScore
    bestWired = copy.deepcopy(wired)
    bestUnwired = copy.deepcopy(unwired)
    bestWiring = tagesschluessel.steckerbrett.wiring
    if self.notify:
     self.notify('{}.shotgunPhase2: bestScore = {}\n{}\n'.format(
      self.__class__.__name__, bestScore, tagesschluessel))
   else:
    tagesschluessel.steckerbrett.wiring = bestWiring
    wired = copy.deepcopy(bestWired)
    unwired = copy.deepcopy(bestUnwired)
   notImproved += 1
   unconnectSrc = random.choice(list(wired))
   connectSrc = random.choice(list(unwired))
   unwired.remove(connectSrc)
   unwired.add(unconnectSrc)
   connectTgt = random.choice(list(unwired))
   tagesschluessel.steckerbrett.clearMark3Setting(unconnectSrc)
   tagesschluessel.steckerbrett.addMark3Setting(connectSrc, connectTgt)
   wired,  unwired = fillWired(tagesschluessel)
  return tagesschluessel

 def exchangePhase2(self, 
   phase1Tagesschluessel : MzEnigma.Tagesschluessel, encodedSpruch : str = '', cycles : int = 5) -> MzEnigma.Tagesschluessel:
  """Hill climbing based on exchange of x-connected ports of a steckerbrett
 
:param phase1Tagesschluessel: tagesschluessel with correct settings for Umkehrwalze, Walzen, Zusatzwalze, and Tageswalzenstellungen(required)
:param encodedSpruch: encrypted message to be attacked (required)
:param cycles: number of exchange cycles
:return: Tagesschlüssel
  """
  assert self.spruchScoring is not None, '{}.exchangePhase2: self.spruchScoring is required'.format(self.__class__.__name__)
  assert phase1Tagesschluessel.steckerbrett is not None, '{}.gilloglyAttackPhase2: steckerbrett is required'.format(self.__class__.__name__)
  tagesschluessel = copy.deepcopy(phase1Tagesschluessel)
  wired = set()
  for n, c in enumerate(tagesschluessel.alphabet):
   if tagesschluessel.steckerbrett.wiring[n] != c:
    wired.add(c)
  if len(wired) <= 2:
   return phase1Tagesschluessel
  bestWiring = tagesschluessel.steckerbrett.wiring
  bestScore = self.spruchScoring.ngramScore(tagesschluessel.decode(encodedSpruch), validChars = tagesschluessel.alphabet)
  for cycle in range(cycles):
   frequencyDict = tagesschluessel.frequencyDict(encodedSpruch, decode = True)
   for freq in sorted(frequencyDict.keys(), reverse = True):
    for c1, _ in frequencyDict[freq]:
     if c1 in wired:
      c1Index = tagesschluessel.steckerbrett.wiring.index(c1)
      if c1 != tagesschluessel.alphabet[c1Index]: 
       for c2 in wired:
        c2Index = tagesschluessel.steckerbrett.wiring.index(c2)
        if c2 != c1 and c2 != tagesschluessel.alphabet[c2Index]:
         actWiring = tagesschluessel.steckerbrett.wiring
         tagesschluessel.steckerbrett.exchangeMark3Setting(c1, c2)
         actScore = self.spruchScoring.ngramScore(tagesschluessel.decode(encodedSpruch), validChars = tagesschluessel.alphabet)
         if actScore > bestScore:
          bestScore = actScore
          bestWiring = tagesschluessel.steckerbrett.wiring
          if self.notify:
           self.notify('{}.exchangePhase2: cycle {}, exchanging {} <-> {}, score = {:.3f}'.format(self.__class__.__name__, cycle + 1, c1, c2, bestScore))
         tagesschluessel.steckerbrett.wiring = actWiring
   tagesschluessel.steckerbrett.wiring = bestWiring
  tagesschluessel.steckerbrett.wiring = bestWiring
  return tagesschluessel

 def gilloglyAttackPhase2(self, 
   phase1Tagesschluessel : MzEnigma.Tagesschluessel, encodedSpruch : str = '', 
      noImprovement : int = 10, cycles : int = 1000,  useSimulatedAnnealing : bool = False) -> MzEnigma.Tagesschluessel:
  """Shotgun hill climbing or simulated annealing attack to derive settings for Steckerbrett
 
:param phase1Tagesschluessel: tagesschluessel with correct settings for Umkehrwalze, Walzen, Zusatzwalze, and Tageswalzenstellungen(required)
:param encodedSpruch: encrypted message to be attacked (required)
:param noImprovement: stop condition after *noImprovement* number of attempts in a cycle
:param cycles: number of cycles with a random steckerbrett configuration
:param useSimulatedAnnealing: use simulated annealing instead of conventional hill climbing
:return: Tagesschlüssel
  """

  if useSimulatedAnnealing:
   self.spruchScoring.setSATemperature(encodedSpruch)
  else:
   self.spruchScoring.setSATemperature('')
  bestScore = 0
  tagesschluessel = copy.deepcopy(phase1Tagesschluessel)
  initialScore = self.spruchScoring.ngramScore(tagesschluessel.decode(encodedSpruch), 1, validChars = tagesschluessel.alphabet)
  bestWiring = tagesschluessel.steckerbrett.wiring
  
  for cycle in range(cycles):
   tagesschluessel = self.shotgunPhase2(self, tagesschluessel, encodedSpruch,  noImprovement)
   actScore = self.spruchScoring.newNgramScore(tagesschluessel.decode(encodedSpruch), bestScore, tagesschluessel.alphabet)
   if self.notify:
    self.notify('{}.gilloglyAttackPhase2/cycle {}: score = {:.3f} (best: {:.3f})'.format(self.__class__.__name__, cycle, actScore, bestScore))
   if actScore > bestScore:
    bestScore = actScore
    bestWiring = tagesschluessel.steckerbrett.wiring
   newSteckerbrett = MzEnigma.Steckerbrett.Mark_3(alphabet = tagesschluessel.alphabet)
   tagesschluessel.steckerbrett.wiring = newSteckerbrett.wiring
  tagesschluessel.steckerbrett.wiring = bestWiring
  if self.notify:
   finalScore = self.spruchScoring.ngramScore(tagesschluessel.decode(encodedSpruch), 1, validChars = tagesschluessel.alphabet)
   self.notify('{}.gilloglyAttackPhase2: score {:.3f} -> {:.3f}'.format(self.__class__.__name__, initialScore, finalScore))
  return tagesschluessel

 def mzAttackPhase2(self, phase1Tagesschluessel : MzEnigma.Tagesschluessel, encodedSpruch : str = '') -> MzEnigma.Tagesschluessel:
  """Systematic hill climbing attack to derive settings for Steckerbrett
Uses the 'ngramScore' for scoring.
 
:param phase1Tagesschluessel: tagesschluessel with correct settings for Umkehrwalze, Walzen, Zusatzwalze, and Tageswalzenstellungen(required)
:param encodedSpruch: encrypted message to be attacked (required)
:return: Tagesschlüssel
  """
  assert self.spruchScoring is not None, '{}.mzAttackPhase2: self.spruchScoring is required'.format(self.__class__.__name__)
  assert phase1Tagesschluessel.steckerbrett is not None, '{}.mzAttackPhase2: steckerbrett is required'.format(self.__class__.__name__)
  tagesschluessel = copy.deepcopy(phase1Tagesschluessel)
  nConnections = 0
  for n, c in enumerate(tagesschluessel.alphabet):
   if tagesschluessel.steckerbrett.wiring[n] != c:
    nConnections += 1
  nConnections = nConnections // 2
  assert nConnections > 0, '{}.mzAttackPhase2: at least one connection in steckerbrett required'.format(self.__class__.__name__)
  tagesschluessel.steckerbrett.wiring = tagesschluessel.alphabet
  monograms = sorted(self.spruchScoring.ngramDict[1].items(), key = lambda c2s: c2s[1], reverse = False)
  assert len(monograms) > 2*nConnections, '{}.mzAttackPhase2: len(monograms) = {} > 2*nConnections = {}'.format(self.__class__.__name__, len(monograms), 2*nConnections)
  initialScore = self.spruchScoring.ngramScore(tagesschluessel.decode(encodedSpruch), 1, validChars = tagesschluessel.alphabet)
  unwired = list(tagesschluessel.alphabet)
  self.spruchScoring.numberOfChars = 1
  connectedChars = ''
  connection = 0
  while connection < nConnections and len(unwired) > 1:
   connectTgt = monograms.pop()[0]
   connectedChars += connectTgt
   if connectTgt not in unwired:
    continue
   maxScore = 0
   maxN = -1
   wiring = tagesschluessel.steckerbrett.wiring
   for n, connectSrc in enumerate(unwired):
    tagesschluessel.steckerbrett.wiring = wiring
    if connectSrc != connectTgt:
     tagesschluessel.steckerbrett.addMark3Setting(connectSrc, connectTgt)
    decodedSpruch = tagesschluessel.decode(encodedSpruch)
    if True:
     actScore = self.spruchScoring.ngramScore(decodedSpruch, 1, validChars = connectedChars)
    else:
     actScore = MzEnigma.SpruchScoring.indexOfCoincidence(decodedSpruch)
    if actScore > maxScore:
     maxScore = actScore
     maxN = n
   tagesschluessel.steckerbrett.wiring = wiring
   connectSrc = unwired[maxN]
   unwired.pop(unwired.index(connectSrc))
   if self.notify:
    self.notify('{}.mzAttackPhase2: connecting {} -> {}, maxScore = {:.3f}'.format(self.__class__.__name__, connectSrc, connectTgt, maxScore))
   if connectSrc != connectTgt:
    tagesschluessel.steckerbrett.addMark3Setting(connectSrc, connectTgt)
    unwired.pop(unwired.index(connectTgt))
    connection += 1
  if self.notify:
   finalScore = self.spruchScoring.ngramScore(tagesschluessel.decode(encodedSpruch), 1, validChars = tagesschluessel.alphabet)
   self.notify('{}.mzAttackPhase2: score {:.3f} -> {:.3f}'.format(self.__class__.__name__, initialScore, finalScore))
  return tagesschluessel




from typing import Optional, Callable, Dict, List, Tuple, Set
import pytest

import itertools
import os, os.path
import re
import random
import string
import MzEnigma

defaultMsg = 'DIEENIGMAISTEINEROTORSCHLUESSELMASCHINEDIEIMZWEITENWELTKRIEGZURVERSCHLUESSELUNGDESNACHRICHTENVERKEHRSDERWEHRMACHTVERWENDETWURDEXAUCHPOLIZEIGEHEIMDIENSTEDIPLOMATISCHEDIENSTESDSSREICHSPOSTUNDREICHSBAHNSETZTENSIEZURGEHEIMENKOMMUNIKATIONEINXTROTZMANNIGFALTIGERVORUNDWAEHRENDDESKRIEGESEINGEFUEHRTERVERBESSERUNGENDERVERSCHLUESSELUNGSQUALITAETGELANGESDENALLIIERTENMITHOHEMPERSONELLENUNDMASCHINELLENAUFWANDDIEDEUTSCHENFUNKSPRUECHENAHEZUKONTINUIERLICHZUENTZIFFERNX'
defaultCrib = 'DIEENIGMAISTEINEROTORSCHLUESSELMASCHINE'

def printFrequencyDict(enigmaSetting : MzEnigma.Tagesschluessel, spruchScoring : MzEnigma.SpruchScoring, actMsg : str) -> None:
 frequencyDict = enigmaSetting.frequencyDict(actMsg)
 for freq in sorted(frequencyDict.keys(), reverse = True):
  msg = ' {} : '.format(freq)
  for c2l in frequencyDict[freq]:
   msg += c2l[0]
  print(msg)
 print(' score = {}'.format(round(spruchScoring.ngramScore(actMsg, 1, validChars = enigmaSetting.alphabet), 3)))

def compareSteckerbrett(phase2EnigmaSetting : MzEnigma.Tagesschluessel, expectedEnigmaSetting : MzEnigma.Tagesschluessel) -> str:
  validResultWiring = ''
  for p2C, eC in zip(phase2EnigmaSetting.steckerbrett.wiring, expectedEnigmaSetting.steckerbrett.wiring):
   if p2C == eC:
    validResultWiring += eC
   else:
    validResultWiring += '.'
  return validResultWiring

def runTuringAttack(enigma : MzEnigma.enigma, name : str, msg : Optional[str] = None, crib : Optional[str] = None, notify : Optional[Callable[[str], None]] = None):
 assert enigma.steckerbrett is not None
 if crib is None:
  crib = msg
 print('Starting engine {}'.format(name))
 if enigma.zusatzwalzen is not None:
  zusatzwalze = enigma.zusatzwalzen[0]
 else:
  zusatzwalze = None
  
 enigmaSetting = MzEnigma.Tagesschluessel(enigma, 
                                                            umkehrwalze = enigma.umkehrwalzen[0], 
                                                            walzen = enigma.walzen[:enigma.numberOfWalzen], 
                                                            tagesWalzenStellungen = ''.join(random.sample(enigma.alphabet, enigma.numberOfWalzen)), 
                                                            steckerbrett = MzEnigma.Steckerbrett('Mark 3', 'ARDCEFTIHXPSVQOKNBLGUMYJWZ'), 
                                                            zusatzwalze = zusatzwalze, 
                                                            notify = None)
 # RVDEAHEJX
 sRange = MzEnigma.TagesschluesselRange(enigma, 
  walzenList = [enigmaSetting.walzen], 
  tagesWalzenStellungenList = [enigmaSetting.tagesWalzenStellungen], 
  umkehrwalzenList = [enigmaSetting.umkehrwalze], 
  zusatzwalzenList = [enigmaSetting.zusatzwalze], 
  spruchScoring = None, 
  notify = None) 
 sRange.notify = notify
 print(enigmaSetting)
 eMsg = enigmaSetting.encode(msg)
 steckerbrettWiring = enigmaSetting.steckerbrett.wiringToDict(False)
 print('text = {}'.format(msg))
 print('crib = {}'.format(crib))
 print('encryptedText = {}'.format(eMsg))
 validCandidates = sRange.turingAttack(encodedSpruch = eMsg, crib = crib, startingPosition = msg.find(crib))
 for tagesschluessel, tws2wiringList in validCandidates:
  print('-----------------\n{}'.format(tagesschluessel))
  for tws, wirings in tws2wiringList:
   print(' - tws = {} ++++++++++++++++++++++++'.format(tws))
   for wiring in wirings:
    missingKeys = list()
    improperKeys = list()
    difference = 'correct'
    for src, tgt in steckerbrettWiring.items():
     if src not in wiring and difference == 'correct':
      missingKeys.append(src)
      difference = 'missing keys'
     elif src in wiring and wiring[src] != tgt:
      improperKeys.append(src)
      difference = 'improper keys'
    if difference == 'missing keys':
     print('  + {}: {}'.format(difference, sorted(missingKeys)))
    elif difference == 'correct':
     print('  + {}'.format(difference))
    elif difference == 'improper keys' and notify:
     print('  + {}: {}'.format(difference, sorted(improperKeys)))

def test_turingAttack(pytestconfig):
 global defaultMsg, defaultCrib
 msg = pytest.helpers.message(pytestconfig)
 if not msg:
  msg = defaultMsg
  crib = defaultCrib
 else:
  crib = None 
  
 pattern = pytestconfig.getoption('component')
 if pattern:
  print('\n--- test_turingAttack ---\n')
  for name, enigma in vars(MzEnigma).items():
   if isinstance(enigma, MzEnigma.Enigma) and re.fullmatch(pattern, name) and enigma.steckerbrett is not None:
    runTuringAttack(enigma, name, msg, crib, 
                           notify = pytest.helpers.notify(pytestconfig))
  
def runGilloglyAttack(enigma : MzEnigma.Enigma, name : str, msg : str, phase2Method : str = 'Mz', notify : Optional[Callable[[str], None]] = None):
 assert enigma.steckerbrett is not None
 print('Starting engine {}'.format(name))
 enigmaSetting = MzEnigma.Tagesschluessel(enigma, notify = None)
 print(enigmaSetting)
 spruchScoring = MzEnigma.SpruchScoring('german', normalize = True, logarithmic = False) 
 eMsg = enigmaSetting.encode(msg)
 print('text = {}'.format(msg))
 print('encryptedText = {}'.format(eMsg))
 print('frequencyDict:')
 printFrequencyDict(enigmaSetting, spruchScoring, msg)
 sRange = MzEnigma.TagesschluesselRange(enigma, 
  walzenList = [enigmaSetting.walzen], 
  tagesWalzenStellungenList = None, 
  umkehrwalzenList = [enigmaSetting.umkehrwalze], 
  zusatzwalzenList = [enigmaSetting.zusatzwalze], 
  spruchScoring = spruchScoring, 
  notify = None) 
 sRange.notify = notify
 phase1EnigmaSetting = sRange.gilloglyAttackPhase1(eMsg)
 phase1Msg = 'phase 1 analysis of settings - Umkehrwalze, Walzen, Zusatzwalze, and Tageswalzenstellungen -'
 if phase1EnigmaSetting != enigmaSetting:
  print('{} of {} failed'.format(phase1Msg, name))
  print('{}\nexpected = \n{}'.format(name, enigmaSetting))
  print('{}\nderived = \n{}'.format(name, phase1EnigmaSetting))
  return
 print('{} succeeded'.format(phase1Msg))
 print('frequencyDict after phase 1:')
 printFrequencyDict(phase1EnigmaSetting, spruchScoring, phase1EnigmaSetting.decode(eMsg))
 if phase2Method == 'Mz':
  phase2EnigmaSetting = sRange.mzAttackPhase2(phase1EnigmaSetting, encodedSpruch = eMsg)
 else:
  phase2EnigmaSetting = sRange.gilloglyAttackPhase2(phase1EnigmaSetting, encodedSpruch = eMsg, noImprovement = 10, cycles = 1000,  useSimulatedAnnealing = False)
 phase2Msg = 'phase 1 analysis of settings - Steckerbrett -'
 if phase2EnigmaSetting.steckerbrett == enigmaSetting.steckerbrett:
  print('{} of {} succeeded'.format(phase2Msg, name))
  return
 print('{} not succeeded, steckerbrett comparison = {}'.format(phase2Msg, compareSteckerbrett(phase2EnigmaSetting, enigmaSetting)))
 if phase2Method == 'Mz':
  print(' ... trying fixes')
  phase2EnigmaSettingNew = sRange.exchangePhase2(phase2EnigmaSetting, encodedSpruch = eMsg)
  print('{} still not succeeded, steckerbrett comparison = {}'.format(phase2Msg, compareSteckerbrett(phase2EnigmaSettingNew, enigmaSetting)))
  if phase2EnigmaSettingNew.steckerbrett == enigmaSetting.steckerbrett:
   print('{} of {} succeeded with fixes'.format(phase2Msg, name))
   return
  decodedMsg = phase2EnigmaSettingNew.decode(eMsg)
 else:
  decodedMsg = phase2EnigmaSetting.decode(eMsg)
 print('Message:\n{} (expected) \n{} (got)'.format(msg, decodedMsg))

def test_gilloglyAttack(pytestconfig):
 msg = pytest.helpers.message(pytestconfig)
 if not msg:
  msg = defaultMsg
 pattern = pytestconfig.getoption('component')
 phase2Method = 'Mz'
 if pattern:
  print('\n--- test_gilloglyAttack ---\n')
  for name, enigma in vars(MzEnigma).items():
   if isinstance(enigma, MzEnigma.Enigma) and re.fullmatch(pattern, name) and enigma.steckerbrett is not None:
    runGilloglyAttack(enigma, name, msg, 
                           phase2Method = phase2Method,  
                           notify = pytest.helpers.notify(pytestconfig))

def runRejewskiAttack(enigma : MzEnigma.Enigma, name : str, usePickleFile : bool = True, notify : Optional[Callable[[str], None]] = None):
 assert enigma.steckerbrett is None
 print('Starting engine {}'.format(name))
 enigmaSetting = MzEnigma.Tagesschluessel(enigma, 
   walzen = enigma.walzen[:enigma.numberOfWalzen], 
   umkehrwalze = enigma.umkehrwalzen[0], notify = None)
 print(enigmaSetting)
 msg = ''.join(random.sample(enigma.alphabet, enigma.numberOfWalzen))
 eMsg = enigmaSetting.encode(msg +  msg)
 print('spruchschluessel = {}'.format(msg))
 print('encryptedText = {}'.format(eMsg))
 sRange = MzEnigma.TagesschluesselRange(enigma, 
  walzenList = [enigmaSetting.walzen], 
  tagesWalzenStellungenList = None, 
  umkehrwalzenList = [enigmaSetting.umkehrwalze], 
  zusatzwalzenList = [enigmaSetting.zusatzwalze], 
  spruchScoring = None, 
  notify = None) 
 sRange.notify = notify
 pickleFile = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'rejewskiCatalog.pickle')
 if usePickleFile and os.path.exists(pickleFile):
  catalog = sRange.loadRejewskiCatalog(pickleFile)
 else:
  if os.path.exists(pickleFile):
   os.remove(pickleFile)
  catalog = sRange.createRejewskiCatalog(pickleFile)
 validCandidates = sRange.rejewskiAttack(catalog, eMsg)
 print('Computation finished, {} candidates found'.format(len(validCandidates)))
 for setting, spruchschluessel in validCandidates:
  found = list()
  if setting == enigmaSetting:
   found.append('setting correct')
  if spruchschluessel == msg:
   found.append('spruchschluessel correct')
  found = ', '.join(found)
  if len(found) == 0:
   found = 'improper'
  print('tagesWalzenStellungen = {}, spruchschluessel = {} -> {} '.format(setting.tagesWalzenStellungen, spruchschluessel, found))

def test_rejewskiAttack(pytestconfig):
 pattern = pytestconfig.getoption('component')
 if pattern:
  print('\n--- test_rejewskiAttack ---\n')
  for name, enigma in vars(MzEnigma).items():
   if isinstance(enigma, MzEnigma.Enigma) and re.fullmatch(pattern, name) and enigma.steckerbrett is None:
    runRejewskiAttack(enigma, name, notify = pytest.helpers.notify(pytestconfig))
                            
if __name__ == "__main__":
 import difflib
 match = difflib.SequenceMatcher(None,'abcd','yycy')
 matchResult = list(match.find_longest_match(0, 4, 0, 4))
 # msg = 'DIEENIGMAISTEINEROTORSCHLUESSELMASCHINEDIEIMZWEITENWELTKRIEGZURVERSCHLUESSELUNGDESNACHRICHTENVERKEHRSDERWEHRMACHTVERWENDETWURDEXAUCHPOLIZEIGEHEIMDIENSTEDIPLOMATISCHEDIENSTESDSSREICHSPOSTUNDREICHSBAHNSETZTENSIEZURGEHEIMENKOMMUNIKATIONEINXTROTZMANNIGFALTIGERVORUNDWAEHRENDDESKRIEGESEINGEFUEHRTERVERBESSERUNGENDERVERSCHLUESSELUNGSQUALITAETGELANGESDENALLIIERTENMITHOHEMPERSONELLENUNDMASCHINELLENAUFWANDDIEDEUTSCHENFUNKSPRUECHENAHEZUKONTINUIERLICHZUENTZIFFERNX'
 # msg = 'DIEENIGMAISTEINEROTORSCHLUESSELMASCHINE'
 msg = "OBERKOMMANDODERWEHRMACHTOBERKOMMANDODERWEHRMACHTOBERKOMMANDODERWEHRMACHT"
 #          'XLNFCBYXCWGAOWURKQJDUBNRQWWXVIKIMPLYQAGFCVUEUZCINVQYSAWLIFEXAAKSPETAJXIY'
 print('plaintext = {}'.format(msg))
 if False:
  # crib = 'NIGMAISTEINEROTORSCHLUESSELMASCHINEDIE'
  crib = msg[:12]
  enigma = MzEnigma.Enigma_D
  name = 'Enigma_D'
  runRejewskiAttack(enigma, name, notify = None)
 elif False:
  # crib = 'NIGMAISTEINEROTORSCHLUESSELMASCHINEDIE'
  crib = msg[:12]
  enigma = MzEnigma.Enigma_M3
  name = 'Enigma_M3'
  runTuringAttack(enigma, name, msg, crib = crib, notify = None)
 else:
  phase2Method = 'Mz'
  enigma = MzEnigma.Enigma_M3
  name = 'Enigma_M3'
  runGilloglyAttack(enigma, name, msg, phase2Method = phase2Method, notify = print)



from __future__ import annotations
from typing import Optional, Union, Callable, List, Tuple, Set, Dict

import copy
import itertools
import random
import csv
import os
import math

import MzEnigma

class SpruchScoring(object):
 """Toolbox for scoring of a spruch. In the background, this object holds
 dictionary with character n-grams. By default, the dictionary with the 
 longest n-gram is used.
 
 :param language: language to be used for ngram scoring (required)
 :param normalize: normalize ngram values with respect to maxScore 
 :param logarithmic: 10*math.log10(ngramDict[score]/minScore)
  """
 def __init__(self, language : str, normalize : bool = True, logarithmic : bool = False):
  self.resPath = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Resources') 
  assert language in ['english', 'german'], "{}: Language {} not supported, use 'english' or 'german'".format(self.__class__.__name__, language)
  self.language = language
  self.normalize = normalize
  self.logarithmic = logarithmic
  self.ngramDict : Dict[int, Dict[str, float]] = dict()
  self.ngramLimits : Dict[int, Dict[str, float]] = dict()
  self.ngramDict[1], self.ngramLimits[1] = self._loadStatistics('{}_monograms.txt'.format(language))
  self.ngramDict[2], self.ngramLimits[2] = self._loadStatistics('{}_bigrams.txt'.format(language))
  self.ngramDict[3], self.ngramLimits[3] = self._loadStatistics('{}_trigrams.txt'.format(language))
  if self.ngramDict[3] is None:
   del self.ngramDict[3]
   del self.ngramLimits[3]
  if self.ngramDict[2] is None:
   del self.ngramDict[2]
   del self.ngramLimits[2]
  if self.ngramDict[1] is None:
   del self.ngramDict[1]
   del self.ngramLimits[1]
  self.numberOfChars = max(self.ngramDict.keys())
  self.saTemperature = 0.0
  self.saThreshold = 0.01
  
 def setSATemperature(self, spruch : str, numberOfChars : int = 0) -> float:
  """Enables the temperature for simulated annealing (used by `newNgramScore`)
 
 :param spruch: spruch of typical length to be scored (required)
 :param numberOfChars: n-gram dictionary to be used (default: longest n-grams) 
  """
  nSpruch = len(spruch)
  if nSpruch == 0:
   self.saTemperature = 0
  elif nSpruch <= 30:
   self.saTemperature = 400.0
  elif nSpruch <= 50:
   self.saTemperature = 400.0 - (400.0 - 315.0) * (nSpruch - 30) / (50 - 30)
  elif nSpruch <= 75:
   self.saTemperature = 315.0 - (315.0 - 240.0) * (nSpruch - 50) / (75 - 50)
  elif nSpruch <= 100:
   self.saTemperature = 240.0 - (240.0 - 220.0) * (nSpruch - 75) / (100 - 75)
  elif nSpruch <= 150:
   self.saTemperature = 220.0 - (220.0 - 200.0) * (nSpruch - 100) / (150 - 100)
  else:
   self.saTemperature = 200.0
  if numberOfChars < 1:
   assert 'numberOfChars' in vars(self), '{}.setSATemperature: no ngrams loaded'.format(self.__class__.__name__)
   numberOfChars = self.numberOfChars
  else:
   assert numberOfChars in self.ngramDict, '{}.setSATemperature: ngramDict[{}] not found'.format(self.__class__.__name__, numberOfChars)
  self.saTemperature /=  10^8
  return self.saTemperature

 def ngramScore(self, spruch : str, numberOfChars : int = 0, validChars : str = MzEnigma.stdAlphabet) -> float:
  """N-gram score of a spruch
 
 :param spruch: spruch to be scored (required)
 :param numberOfChars: n-gram dictionary to be used (default: longest n-grams) 
 :param validChars: alphabet of the spruch
 :returns: N-gram score
  """
  _score = 0
  if numberOfChars < 1:
   assert 'numberOfChars' in vars(self), '{}.ngramScore: no ngrams loaded'.format(self.__class__.__name__)
   numberOfChars = self.numberOfChars
  else:
   assert numberOfChars in self.ngramDict, '{}.ngramScore: ngramDict[{}] not found'.format(self.__class__.__name__, numberOfChars)
  nChars = len(list(self.ngramDict[numberOfChars].keys())[0])
  for c in validChars:
   assert c in self.ngramDict[numberOfChars], '{}.ngramScore: {} not in ngramDict[{}]'.format(self.__class__.__name__, c, numberOfChars)
  for n in range(len(spruch) - nChars + 1):
   ngram = spruch[n: n + nChars]
   for c in ngram:
    if c in validChars:
     _score += self.ngramDict[numberOfChars][ngram]
     break
  return _score/(len(spruch) - nChars + 1)

 def newNgramScore(self, spruch : str, currentScore : float, validChars : str = MzEnigma.stdAlphabet) -> float:
  """Compares the N-gram score of a spruch with the currentScore. 
  Uses simulated if enabled.
 
 :param spruch: spruch to be scored (required)
 :param currentScore: current score (required)
 :param validChars: alphabet of the spruch
 :returns: N-gram score
  """
  score = self.ngramScore(spruch, validChars = validChars)
  diffScore = score - currentScore
  if diffScore >= 0:
   return score
  if self.saTemperature <= 0:
   return currentScore
  diffScore /= self.saTemperature
  prob = math.exp(diffScore)
  if prob <= random.uniform(self.saThreshold, 1.0):
   score = currentScore
  return score

 @staticmethod
 def indexOfCoincidence(spruch : str, validChars : str = MzEnigma.stdAlphabet) -> float:
  """index of coincidence of a spruch,
  i.e. relative frequency of characters independent of the underlying language 
 
 :param spruch: spruch to be scored (required)
 :param validChars: alphabet of the spruch
 :returns: index of coincidence
  """
  nAlphabet = len(validChars)
  nSpruch = len(spruch)
  iocList = nAlphabet * [0.0]
  for c in spruch:
   iocList[validChars.index(c)] += 1
  ioc = 0
  for n in range(nAlphabet):
   ioc += iocList[n] * (iocList[n] - 1)
  return ioc/(nSpruch * (nSpruch-1))

 def _loadStatistics(self, statfile : str) -> Dict[str, float]:
  ngramDict : Dict[str, Union[float, int]] = dict()
  alphabet = set(MzEnigma.stdAlphabet) | set(MzEnigma.sgsAlphabet)
  actPath = os.path.join(self.resPath, statfile)
  if not os.path.exists(actPath):
   return None, None
  with open(actPath, newline='') as csvfile:
   for ngram, nDetected in csv.reader(csvfile, delimiter=' '):
    isInside = True
    for c in ngram:
     isInside &= c in alphabet
    if isInside:
     nDetected = float(nDetected)
     assert nDetected > 0
     ngramDict[ngram] = nDetected
  maxValue = max(ngramDict.values())
  minValue = min(ngramDict.values())
  if self.logarithmic:
   ngramDict.update({ngram : 10*math.log10(ngramDict[ngram]/minValue) for ngram in ngramDict.keys()})
  elif self.normalize:
   ngramDict.update({ngram : ngramDict[ngram]/maxValue for ngram in ngramDict.keys()})
  else:
   minValue /= maxValue
   maxValue = 1
  return ngramDict, (minValue, maxValue)

 def __repr__(self) -> str:
  return '{}\n language: {}\n normalized: {}\n logarithmic: {}'.format(self.__class__.__name__, self.language, self.normalize, self.logarithmic)

class Enigma(object):
 """Represents an Enigma engine with all available accessories. 
    If no umkehrwalzen are defined, the encryption is not symmetric any more !
 
 :param model: model of the Enigma (required)
 :param umkehrwalzen: available objects of type 'Umkehrwalze' (if any)
 :param walzen: available objects of type 'Walze' (at least 1 required)
 :param numberOfWalzen: number of  objects of type 'Walze' to be inserted
 :param steckerbrett: object of type 'Steckerbrett', if any
 :param zusatzwalzen: available objects of type 'Zusatzwalze', if any
 :param notify: notification function (e.g. print)
  """
 def __init__(self, 
  model  : str = '', 
  umkehrwalzen : List[MzEnigma.Umkehrwalze] = None, 
  walzen : List[MzEnigma.Walze] = None, 
  numberOfWalzen : int = 0, 
  steckerbrett : Optional[MzEnigma.Steckerbrett] = None, 
  zusatzwalzen : List[MzEnigma.Zusatzwalze] = None) -> None:
  assert model, '{}: At least 1 character in model required'.format(self.__class__.__name__)
  self._model = model
  assert walzen, '{}: At least 1 walze required'.format(self.__class__.__name__)
  alphabet = walzen[0].alphabet
  assert all(v.alphabet == alphabet for v in walzen), '{}: Alphabets in all walzen must match'.format(self.__class__.__name__)
  self._walzen = walzen
  if umkehrwalzen:
   assert all(v.alphabet == alphabet for v in umkehrwalzen), '{}: Alphabets in all umkehrwalzen must match'.format(self.__class__.__name__)
  self._umkehrwalzen = umkehrwalzen
  assert numberOfWalzen > 0 and numberOfWalzen <= len(walzen), '{}: 0 < numberOfWalzen < {}  required'.format(self.__class__.__name__, len(walzen))
  self._numberOfWalzen = numberOfWalzen
  if steckerbrett:
   assert steckerbrett.alphabet == alphabet, '{}: Alphabets in steckerbrett must match'.format(self.__class__.__name__)
  self._steckerbrett = steckerbrett
  if zusatzwalzen:
   assert all(v.alphabet == alphabet for v in zusatzwalzen), '{}: Alphabets in all zusatzwalzen must match'.format(self.__class__.__name__)
  self._zusatzwalzen = zusatzwalzen

 @staticmethod
 def validCribPositions(encodedSpruch : str = '', crib : str = '') -> List[int]:
  """This function delivers all positions where no crib letter encrypts as itself
  
:param encodedSpruch: message to be decoded
:param crib: substring to be examined
:returns: list of valid positions relative to the encodedSpruch
   """
  assert len(encodedSpruch) > 0, 'validCribPositions: encodedSpruch is missing'
  assert len(crib) > 0, 'validCribPositions: crib is missing'
  assert len(crib) <= len(encodedSpruch), 'validCribPositions: len(encodedSpruch) == {} < len(crib) == {}'.format(len(encodedSpruch), len(crib))

  validPosList = list()
  for startPos in range(0, len(encodedSpruch) - len(crib) + 1):
   # match = difflib.SequenceMatcher(None, encodedSpruch[startPos:], crib)
   isMatched = False
   for ps, pc in zip(encodedSpruch[startPos:], crib):
    if ps == pc:
     isMatched = True
     break
   if not isMatched:
    validPosList.append(startPos)
  return validPosList
  
 @property
 def alphabet(self) -> str:
  """
  :getter: Returns the alphabet of the engine
  :setter: None
  """
  return self._walzen[0].alphabet

 @property
 def model(self) -> str:
  """
  :getter: Returns the name of the engine
  :setter: None
  """
  return self._model

 @property
 def umkehrwalzen(self) -> List[MzEnigma.Umkehrwalze]:
  """
  :getter: Returns a list of Umkehrwalzen
  :setter: None
  """
  return self._umkehrwalzen

 @property
 def walzen(self) -> List[MzEnigma.Walze]:
  """
  :getter: Returns a list of Walzen
  :setter: None
  """
  return self._walzen

 @property
 def numberOfWalzen(self) -> int:
  """
  :getter: Returns a number of Walzen built-in at the engine's operation
  :setter: None
  """
  return self._numberOfWalzen

 @property
 def steckerbrett(self) -> Optional[int]:
  """
  :getter: Returns the steckerbrett if any
  :setter: None
  """
  return self._steckerbrett

 @property
 def zusatzwalzen(self) -> Optional[List[MzEnigma.Zusatzwalze]]:
  """
  :getter: Returns a list of Zusatzwalzen if any
  :setter: None
  """
  return self._zusatzwalzen
  
 def __eq__(self, enigma : Enigma) -> bool:
  return self._model == enigma.model \
     and self._umkehrwalzen == enigma.umkehrwalzen \
     and self._walzen == enigma.walzen \
     and self._numberOfWalzen == enigma.numberOfWalzen \
     and self._steckerbrett == enigma.steckerbrett \
     and self._zusatzwalzen == enigma.zusatzwalzen 

 def __repr__(self) -> str:
  content = 'class: {}\nmodel: {}\nalphabet: {}\nUmkehrwalzen:'.format(self.__class__.__name__, self._model, self.alphabet)
  for walze in self._umkehrwalzen:
   for line in walze.__repr__().split('\n'):
    if 'class:' not in line and 'alphabet:' not in line:
     content += '\n ' + line
  content += '\nnumber of Walzen: {}\nWalzen:'.format(self._numberOfWalzen)
  for walze in self._walzen:
   for line in walze.__repr__().split('\n'):
    if 'class:' not in line and 'alphabet:' not in line:
     content += '\n ' + line
  if self._steckerbrett:
   content += '\nSteckerbrett:'
   for line in self._steckerbrett.__repr__().split('\n'):
    if 'class:' not in line and 'alphabet:' not in line:
     content += '\n ' + line
  if self._zusatzwalzen:
   content += '\nZusatzwalzen:'
   for walze in self._zusatzwalzen:
    for line in walze.__repr__().split('\n'):
     if 'class:' not in line and 'alphabet:' not in line:
      content += '\n ' + line
  return content

class Tagesschluessel(object):
 """Represents the Tagesschluessel of an Enigma
 
:param enigma: model of the Enigma (required)
:param umkehrwalze: 'Umkehrwalze' (if undefined, randomly selected)
:param walzen: List of type 'Walze' (if undefined, randomly selecte(if undefined, randomly selected)
:param zusatzwalze: 'Zusatzwalze' (if undefined, randomly selected)
:param blank: replacement character for a blank
:param notify: notification function (e.g. print)
  """
 def __init__(self,
  enigma : Optional[Enigma] = None,  
  umkehrwalze : Optional[MzEnigma.Umkehrwalze] = None, 
  walzen : Optional[List[MzEnigma.Walze]] = None,
  tagesWalzenStellungen : Optional[str] = None, 
  steckerbrett : Optional[MzEnigma.Steckerbrett] = None, 
  zusatzwalze : Optional[MzEnigma.Zusatzwalze] = None, 
  blank : str = '',  
  notify : Optional[Callable[[str], None]] = None) -> None:
  self.notify = notify
  assert enigma, '{}: Enigma required'.format(self.__class__.__name__)
  self.enigma =enigma
  self.blank = blank
  alphabet = self.enigma.alphabet
  if steckerbrett is None and enigma.steckerbrett:
   self.steckerbrett = copy.deepcopy(enigma.steckerbrett)
   self.steckerbrett.notify = self.notify
  else:
   self.steckerbrett = steckerbrett
  lastWalze = self.steckerbrett
  if not walzen:
   walzen = random.sample(enigma.walzen, enigma.numberOfWalzen)
  assert all(v.alphabet == alphabet for v in walzen), '{}: Alphabet in all walzen must match enigma'.format(self.__class__.__name__)
  self.walzen = list()
  for _walze in walzen:
   walze = copy.deepcopy(_walze)
   walze.notify = self.notify
   self.walzen.append(walze)
   if lastWalze:
    lastWalze.nextComponent = walze
    walze.prevComponent = lastWalze
   lastWalze = walze
  if self.steckerbrett:
   self.firstComponent = self.steckerbrett
  else:
   self.firstComponent = self.walzen[0]
  if tagesWalzenStellungen:
   assert len(tagesWalzenStellungen) == len(walzen), '{}: Walzenstellung, number of characters must match the number of walzen'.format(self.__class__.__name__)
   assert all(v in alphabet for v in tagesWalzenStellungen), '{}: Walzenstellung, all characters must be in alphabet'.format(self.__class__.__name__)
  else:
   tagesWalzenStellungen = ''.join(random.sample(alphabet, enigma.numberOfWalzen))
  self.tagesWalzenStellungen = tagesWalzenStellungen
  if zusatzwalze:
   assert zusatzwalze.alphabet == alphabet, '{}: Alphabet in zusatzwalze must match enigma'.format(self.__class__.__name__)
  elif enigma.zusatzwalzen is not None:
   zusatzwalze = random.sample(enigma.zusatzwalzen, 1)[0]
  if zusatzwalze:
   self.zusatzwalze = copy.deepcopy(zusatzwalze)
   self.zusatzwalze.notify = self.notify
   lastWalze.nextComponent = self.zusatzwalze
   self.zusatzwalze.prevComponent = lastWalze
   lastWalze = self.zusatzwalze
  else:
   self.zusatzwalze = None
  if umkehrwalze:
   assert umkehrwalze.alphabet == alphabet, '{}: Alphabet in umkehrwalze must match enigma alphabet'.format(self.__class__.__name__)
  elif len(enigma.umkehrwalzen) > 0:
   umkehrwalze = random.sample(enigma.umkehrwalzen, 1)[0]
  else:
   self.umkehrwalze = None
  if umkehrwalze:
   self.umkehrwalze = copy.deepcopy(umkehrwalze)
   self.umkehrwalze.notify = self.notify
   lastWalze.nextComponent = self.umkehrwalze
   self.umkehrwalze.nextComponent = lastWalze

 @classmethod
 def changeWalzen(cls, 
   currentTagesschluessel : Tagesschluessel, 
   umkehrwalze : Optional[MzEnigma.Umkehrwalze] = None, 
   walzen : List[MzEnigma.Walze] = None, 
   tagesWalzenStellungen : Optional[str] = None, 
   zusatzwalze : Optional[MzEnigma.Zusatzwalze] = None) -> None:
  """Changes certain elements of a Tagesschluessel
 
:param currentTagesschluessel: current Tagesschluessel (required)
:param tagesWalzenStellungen: changed Tageswalzenstellungen to be used (if undefined, randomly selected)
:param umkehrwalze: 'Umkehrwalze' (if undefined, used from currentTagesschluessel)
:param walzen: List of type 'Walze' (if undefined, used from currentTagesschluessel)
:param zusatzwalze: 'Zusatzwalze' (if undefined, used from currentTagesschluessel)
:returns: Tagesschluessel object
   """
  if umkehrwalze is None:
   umkehrwalze = currentTagesschluessel.umkehrwalze
  if walzen is None:
   walzen = currentTagesschluessel.walzen
  if zusatzwalze is None:
   zusatzwalze = currentTagesschluessel.zusatzwalze
  return cls(
   enigma = currentTagesschluessel.enigma,  
   umkehrwalze = umkehrwalze, 
   steckerbrett = currentTagesschluessel.steckerbrett, 
   walzen = walzen, 
   tagesWalzenStellungen = tagesWalzenStellungen, 
   zusatzwalze = zusatzwalze, 
   blank = currentTagesschluessel.blank,  
   notify = currentTagesschluessel.notify)

 def findPatterns(self, substringList : List[str] = [''], encodedSpruchList : List[str] = []) -> Dict[str, Set[str]]:
  """Find a list of substringList for any tagesWalzenStellungen at any positions
 
:param substringList: list of substrings
:param encodedSpruchList: list of messages to be searched for substrings
:return: dict(substring, set of tagesWalzenStellungen)
  """
  assert len(encodedSpruchList) > 0
  maxSpruchlength = max(encodedSpruchList, key=len)
  charSet = set()
  for substring in substringList:
   for c in substring:
    assert c in self.alphabet, '{}: Character {} not in {}'.format(self.__class__.__name__, c, self.enigma.alphabet)
    charSet.add(c)
  savedTagesWalzenStellungen = self.tagesWalzenStellungen 
  patternDict : Dict[str, Set[str]] = dict()
  for tagesWalzenStellungenTuple in itertools.permutations(self.alphabet, len(savedTagesWalzenStellungen)):
   self.tagesWalzenStellungen = ''.join(tagesWalzenStellungenTuple)
   charEncodingDict = dict()
   for c in charSet:
    charEncodingDict[c] = self.encode(maxSpruchlength * c)
   for encodedSpruch in encodedSpruchList:
    lSpruch = len(encodedSpruch)
    for substring in substringList:
     isInSpruch = False
     ls = lSpruch - len(substring)
     for n, (c0, cExp) in enumerate(zip(encodedSpruch[:ls], charEncodingDict[substring[0]][:ls])):
      if c0 == cExp:
       for m, cm in enumerate(substring[1:]):
        nm = n + m + 1
        if cm == charEncodingDict[substring[m]][nm]:
         if m == len(substring) - 2:
          isInSpruch = True
          break
     if isInSpruch:
      if substring not in patternDict:
       patternDict[substring] = list()
      patternDict[substring].add(self.tagesWalzenStellungen)
  return patternDict

 def findDoublets(self, first : int = 0, second : int = 3, all : bool = True) -> Union[List[List[str]], List[str]]:
  """Find doublets of any character at fixed positions
 
:param first: first position
:param second: second position (> first)
:return: IF ALL it returns a list of doublets for every position ELSE it returns a list of doublets
  """
  assert first >= 0 and second > first
  allDoubletsList = list()
  if all:
   shift = second - first
  else:
   shift = 1
  ecList = list()
  for nc, c in enumerate(self.alphabet):
   ecList.append(list())
   for cs in self.encode((second + shift) * c):
    ecList[-1].append(cs)
  for ns in range(shift):
   doubletsList = list()
   for n, c in enumerate(self.alphabet):
    doubletsList.append(ecList[n][first+ns] + ecList[n][second+ns])
   allDoubletsList.append(doubletsList)
  if not all:
   return doubletsList
  return allDoubletsList

 def findTagesWalzenStellungen(self, rejewskiList : List[Dict[str, Set[str]]], twiceEncodedSpruchWalzenStellungen : Union[List[str], str] = '') -> List[str]:
  if isinstance(twiceEncodedSpruchWalzenStellungen, str):
   twiceEncodedSpruchWalzenStellungen = [twiceEncodedSpruchWalzenStellungen]
  tagesWalzenStellungen = None
  for tesws in twiceEncodedSpruchWalzenStellungen:
   shift = len(tesws) // 2
   assert shift == len(rejewskiList), '{}.findTagesWalzenStellungen: len(twiceEncodedSpruchWalzenStellungen) = {} == 2*len(rejewskiList) = {}'.format(self.__class__.__name__, shift, 2*len(rejewskiList))
   for n, doubletsDict in enumerate(rejewskiList):
    pair = tesws[n] + tesws[n+shift]
    assert pair in rejewskiList[n].keys(), '{}.findTagesWalzenStellungen: {} not a key of rejewskiList[{}]'.format(self.__class__.__name__, pair, n)
    if tagesWalzenStellungen is None:
     tagesWalzenStellungen = rejewskiList[n][pair]
    else:
     tagesWalzenStellungen &= rejewskiList[n][pair]
  return sorted(tagesWalzenStellungen)

 def encode(self, spruch : str, spruchWalzenStellungen : Optional[str] = None) -> str:
  """Runs forward through the enigma
     If the spruchWalzenStellungen is defined, 
     - it is encoded twice using the tagesWalzenStellungen
     - the spruch is then encoded using the spruchWalzenStellungen
     otherwise the spruch is encoded using the tagesWalzenStellungen

:param spruch: message to be encoded
:param spruchWalzenStellungen: see above
:returns: encoded spruch
  """
  for walze, ringstellung in zip(self.walzen, self.tagesWalzenStellungen):
   walze.ringstellung = ringstellung
  out = str()
  if spruchWalzenStellungen:
   assert len(spruchWalzenStellungen) == len(self.tagesWalzenStellungen), '{}.encode: Walzenstellung, number of characters must match the number of walzen'.format(self.__class__.__name__)
   assert all(v in self.alphabet for v in spruchWalzenStellungen), '{}.encode: Walzenstellung, all characters must be in alphabet'.format(self.__class__.__name__)
   for n in range(2):
    for c in spruchWalzenStellungen:
     out += self._encodeLetter(c)
   for walze, ringstellung in zip(self.walzen, spruchWalzenStellungen):
    walze.ringstellung = ringstellung
   if self.notify is not None:
    self.notify('{}.encode/spruchWalzenStellungen: {} => {}'.format(self.__class__.__name__, spruchWalzenStellungen, out[:len(spruchWalzenStellungen)]))
  for c in spruch.upper():
   if c == ' ':
    c = self.blank
   if len(c) == 1:
    out += self._encodeLetter(c)
  if self.notify is not None:
   self.notify('{}.encode: {} => {}'.format(self.__class__.__name__, spruch, out))
  for walze, ringstellung in zip(self.walzen, self.tagesWalzenStellungen):
   walze.ringstellung = ringstellung
  return out

 def encodeMatrix(self, msgLen : int) -> List[List[str]]:
  """Runs forward through the enigma using all characters of the alphabet

:param msgLen: length of messages
:returns: list of encoded messages
  """
  ecList = list()
  for nc, c in enumerate(self.enigma.alphabet):
   ecList.append(list())
   for cs in self.encode(msgLen * c):
    ecList[-1].append(self.enigma.alphabet.index(cs))
  return ecList

 def decode(self, encodedSpruch : str, useSpruchWalzenStellungen : bool = False) -> str:
  """Runs backward through the enigma
     If useSpruchWalzenStellungen, 
     - the spruchWalzenStellung is expected twice at the beginning
     - the spruch is then decoded using the spruchWalzenStellungen
     otherwise the spruch is decoded using the tagesWalzenStellungen

:param encodedSpruch: message to be decoded
:param useSpruchWalzenStellungen: see above
:returns: decoded spruch
  """
  for walze, ringstellung in zip(self.walzen, self.tagesWalzenStellungen):
   walze.ringstellung = ringstellung
  if useSpruchWalzenStellungen:
   spruchWalzenStellungen = str()
   ls = len(self.tagesWalzenStellungen)
   for c in encodedSpruch[:2*ls]:
    spruchWalzenStellungen += self._encodeLetter(c)
   assert spruchWalzenStellungen[:ls] == spruchWalzenStellungen[ls:], '{}.decode: Inconsistent Spruchschlüssel {}'.format(self.__class__.__name__, spruchWalzenStellungen)
   if self.notify is not None:
    self.notify('{}.decode/spruchWalzenStellungen: {} => {}'.format(self.__class__.__name__, encodedSpruch[:ls], spruchWalzenStellungen[:ls]))
   for walze, ringstellung in zip(self.walzen, spruchWalzenStellungen[:ls]):
    walze.ringstellung = ringstellung
   encodedSpruch = encodedSpruch[2*ls:]
  out = str()
  for c in encodedSpruch:
   out += self._encodeLetter(c)
  for walze, ringstellung in zip(self.walzen, self.tagesWalzenStellungen):
   walze.ringstellung = ringstellung
  return out
     
 def _encodeLetter(self, c : str) -> str:
  """Encodes a single character and steps the first Walze
  
:param c: single character
:returns: encoded character
  """
  assert len(c) == 1, '{}.encodeLetter {}: len({}) != 1'.format(self.__class__.__name__, self.name, c)
  assert c in self.enigma.alphabet, '{}.encodeLetter: letter ({}) not in alphabet'.format(self.__class__.__name__, c)
  self.walzen[0].step()
  return self.firstComponent.encode(c, forward = True)
  
 def chain(self) -> List[MzEnigma.Component]:
  """Creates a chain of visited components

:returns: chain list
  """
  chainList = self.firstComponent.chain(forward = True)
  return chainList
  
 def debugEncode(self, c : str ) -> str:
  """Shows the transformation of a character after each component
 
:param c: single character
:return: text built up from the characters after each component
  """
  assert len(c) == 1, '{}.debugEncode: len({}) != 1'.format(self.__class__.__name__, c)
  assert c in self.enigma.alphabet, '{}.debugEncode: Character {} out of range'.format(self.__class__.__name__, c)
  for walze, ringstellung in zip(self.walzen, self.tagesWalzenStellungen):
   walze.ringstellung = ringstellung
  output = ''
  actC = c
  for component, forward in self.chain():
   actC = component.encode(actC, forward, componentOnly = True)
   output += actC
  return output
   
 def frequencyDict(self, spruch : str, decode : bool = True) -> Dict[int, Set[Tuple[str, Set[str]]]]:
  """Creates a frequency dictionary of a message
 
:param spruch: message to be analysed (required)
:param decode: use decoced message instead of the original message
:return: Dict[frequency, Set[Tuple[char, Tuple[occurrency]]
  """
  if decode:
   spruch = self.decode(spruch)
  assert len(spruch) > 0
  c2sDict = dict()
  for n, sc in enumerate(spruch):
   if sc not in c2sDict:
    c2sDict[sc] = set()
   c2sDict[sc].add(n)
  l2csDict = dict()
  for c, nSet in c2sDict.items():
   l = len(nSet)
   if l not in l2csDict:
    l2csDict[l] = set()
   l2csDict[l].add((c, tuple(sorted(nSet))))
  return l2csDict
  
 @property
 def alphabet(self) -> str:
  """
  :getter: Returns the alphabet of the engine
  :setter: None
  """
  return self.enigma.alphabet

 def __eq__(self, tagesschluessel : Tagesschluessel) -> bool:
  return self.enigma == tagesschluessel.enigma \
     and self.tagesWalzenStellungen == tagesschluessel.tagesWalzenStellungen \
     and self.umkehrwalze == tagesschluessel.umkehrwalze \
     and self.walzen == tagesschluessel.walzen \
     and self.steckerbrett == tagesschluessel.steckerbrett \
     and self.zusatzwalze == tagesschluessel.zusatzwalze

 def __repr__(self) -> str:
  content = 'class: {}\nmodel: {}\nalphabet: {}'.format(self.__class__.__name__, self.enigma.model, self.alphabet)
  for walze, ringstellung in zip(self.walzen, self.tagesWalzenStellungen):
   content += '\nWalze {}, Stellung: {}, wiring: {}, notches: {}'.format(walze.name, ringstellung, walze.wiring, walze.notches)
  content += '\nUmkehrwalze: {}, wiring: {}'.format(self.umkehrwalze.name, self.umkehrwalze.wiring)
  if self.steckerbrett:
   content += '\nSteckerbrett {}, wiring: {}'.format(self.steckerbrett.name, self.steckerbrett.wiring)
  if self.zusatzwalze:
   content += '\nZusatzwalze: {}, wiring: {}'.format(self.zusatzwalze.name, self.zusatzwalze.wiring)
  return content




import shutil
import sys
import os, os.path

# Try using setuptools first, if it's installed
from setuptools import setup
# from packaging import version

package = 'MzEnigma'
fileDirectory = os.path.dirname(os.path.abspath(__file__))
packageDirectory = os.path.join(fileDirectory, package)
sys.path.insert(0, fileDirectory)

with open(os.path.join(packageDirectory,'readme.rst'), 'r', encoding = 'utf-8') as f:
 long_description = f.read()
 
import MzEnigma
pkgVersion = MzEnigma.__version__

pythonVersion = sys.version_info[0] + 0.1 * sys.version_info[1]
if pythonVersion < 3.9:
 networkx = 'networkx>=2.2'
else:
 networkx = 'networkx>=2.5'

# Required to ensure a clean environment
shutil.rmtree(os.path.join(fileDirectory, 'build'), ignore_errors = True)

# Need to add all dependencies to setup as we go!
setup(name = package,
  url = 'https://github.com/ReinhardM-dev/MzEnigma', 
  project_urls={ 'Documentation': 'https://reinhardm-dev.github.io/MzEnigma' }, 
  version = pkgVersion,
  packages = [package],
  options={'bdist_wheel':{'universal':True}},
  package_data = {package: ['*.txt', '*.gpl3', '*.rst', 'Resources/*.txt'] }, 
  description = 'A Simulator for Enigma operation and analysis',
  long_description = long_description, 
  long_description_content_type="text/x-rst",
  author  ='Reinhard Maerz',
  python_requires = '>=3.7', 
  install_requires = [ networkx ],
  setup_requires=['wheel'], 
  classifiers = [
    'Programming Language :: Python', 
    'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3 :: Only',
    'Development Status :: 4 - Beta', 
    'Natural Language :: English', 
    'Topic :: Security :: Cryptography'])

